import os
import sys
import re
from os import path, listdir
from platform import system
from importlib import util
from logging import getLogger
from inspect import getmembers, isclass, isfunction
from typing import Iterator

from ndu_gate_camera.utility import constants, string_helper

log = getLogger("service")


class NDUUtility:
    # Buffer for connectors/converters
    # key - class name
    # value - loaded class
    loaded_runners = {}

    @staticmethod
    def check_and_import(extension_type, module_name, package_uuids=None, extension_folder=None):
        if NDUUtility.loaded_runners.get(extension_type + module_name) is None:
            file_dir = path.dirname(path.dirname(__file__))
            extensions_paths = []
            if extension_folder is not None:
                extensions_paths.append((os.path.join(extension_folder, extension_type.lower())).replace('/', path.sep))

            if system() == "Windows":
                extensions_paths.append(
                    path.abspath(file_dir + '/runners/'.replace('/', path.sep) + extension_type.lower()))
            else:
                extensions_paths.append('/var/lib/ndu_gate/runners/'.replace('/', path.sep) + extension_type.lower())
                extensions_paths.append(
                    path.abspath(file_dir + '/runners/'.replace('/', path.sep) + extension_type.lower()))

            #dev_runner_folder = path.abspath(file_dir + '/../runners/'.replace('/', path.sep) + extension_type.lower())
            #if path.isdir(dev_runner_folder):
            #    extensions_paths.append(dev_runner_folder)
            # extensions_paths.append(path.abspath(file_dir + '/../runners/'.replace('/', path.sep) + extension_type.lower()))

            if package_uuids and len(package_uuids) > 0:
                for uuid in package_uuids:
                    if uuid and type(uuid) is str:
                        extensions_paths.append('/var/lib/ndu_gate/runners/'.replace('/', path.sep) + 'Pack_' + uuid)

            try:
                handled = []
                for extension_path in extensions_paths:
                    if extension_path in handled:
                        continue
                    handled.append(extension_path)
                    if path.exists(extension_path):
                        log.info("Looking for %s in %s", module_name, extension_path)
                        for file in listdir(extension_path):
                            if file.startswith('__') or not file.endswith('.py'):
                                continue
                            try:
                                module_spec = util.spec_from_file_location(module_name,
                                                                           extension_path + path.sep + file)
                                log.debug(module_spec)

                                if module_spec is None:
                                    log.error('Module not found : %s', module_name)
                                    continue

                                module = util.module_from_spec(module_spec)
                                log.info(str(module))
                                module_spec.loader.exec_module(module)
                                for extension_class in getmembers(module, isclass):
                                    if module_name in extension_class:
                                        log.info("Import %s from %s", module_name, extension_path)
                                        NDUUtility.loaded_runners[extension_type + module_name] = [extension_class[1],
                                                                                                   extension_path]
                                        log.info("Total runners : %s", len(NDUUtility.loaded_runners))
                                        return NDUUtility.loaded_runners[extension_type + module_name]
                            except Exception as ie:
                                log.error(ie)
                                continue
                    else:
                        log.error("Import %s failed, path doesn't exist: %s", module_name, extension_path)
            except Exception as e:
                log.exception(e)
        else:
            log.info("Class %s found in NDUUtility buffer.", module_name)
            return NDUUtility.loaded_runners[extension_type + module_name]

    @staticmethod
    def get_methods(cls_):
        methods = getmembers(cls_, isfunction)
        return dict(methods)

    @staticmethod
    def has_method(cls_, name):
        methods = getmembers(cls_, isfunction)
        return name in dict(methods)

    @staticmethod
    def install_package(package, version="upgrade"):
        from sys import executable
        from subprocess import check_call, CalledProcessError
        result = False
        if version.lower() == "upgrade":
            try:
                result = check_call([executable, "-m", "pip", "install", package, "--upgrade", "--user"])
            except CalledProcessError:
                try:
                    result = check_call([executable, "-m", "pip", "install", package, "--upgrade"])
                except Exception as e:
                    log.error(e)
        else:
            from pkg_resources import get_distribution
            current_package_version = None
            try:
                current_package_version = get_distribution(package)
            except Exception as e:
                log.error(e)
                pass
            if current_package_version is None or current_package_version != version:
                installation_sign = "==" if ">=" not in version else ""
                try:
                    result = check_call(
                        [executable, "-m", "pip", "install", package + installation_sign + version, "--user"])
                except CalledProcessError:
                    try:
                        result = check_call([executable, "-m", "pip", "install", package + installation_sign + version])
                    except Exception as e:
                        log.error(e)

        return result

    @staticmethod
    def is_url_valid(url) -> bool:
        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)

        return re.match(regex, url) is not None

    @staticmethod
    def is_debug_mode():
        get_trace = getattr(sys, 'gettrace', None)
        return get_trace is not None

    # sadece preview/debug için kullanılmalıdır.
    @staticmethod
    def debug_conv_turkish(class_name):
        if class_name == "person":
            class_name = "insan"

        elif class_name == "car":
            class_name = "otomobil"
        elif class_name == "bicycle":
            class_name = "bisiklet"
        elif class_name == "motorbike":
            class_name = "motosiklet"
        elif class_name == "truck":
            class_name = "kamyon"
        elif class_name == "bus":
            class_name = "otobus"

        elif class_name == "helmet":
            class_name = "baret takiyor"
        elif class_name == "no helmet":
            class_name = "baret takmiyor"

        elif class_name == "mask":
            class_name = "maske takiyor"
        elif class_name == "no mask":
            class_name = "maske takmiyor"

        elif class_name == "good_speed":
            class_name = "Hız"

        class_name = class_name.replace("sec.", "sn.")

        return class_name

    # sadece preview/debug için kullanılmalıdır.
    @staticmethod
    def debug_replace_tur_chars(txt):
        return txt \
            .replace("İ", "I").replace("Ğ", "G").replace("Ü", "U").replace("Ş", "S").replace("Ç", "C").replace("Ö",
                                                                                                               "O").replace(
            "Â", "A") \
            .replace("ı", "i").replace("ğ", "g").replace("ü", "u").replace("ş", "s").replace("ç", "c").replace("ö",
                                                                                                               "o").replace(
            "â", "a")

    # returns (class_name, score, rect) from extra_data. Score and rect can be None!
    @staticmethod
    def enumerate_results(extra_data, class_name_filters=None, use_wildcard=False, return_item=False) -> Iterator[
        tuple]:
        results = extra_data.get(constants.EXTRA_DATA_KEY_RESULTS, None)
        if results is not None:
            for _runner_name, result in results.items():
                # return NDUUtility.enumerate_result_items(result, class_name_filters=class_name_filters, use_wildcard=use_wildcard, return_item=return_item)
                for item in NDUUtility.enumerate_result_items(result, class_name_filters=class_name_filters,
                                                              use_wildcard=use_wildcard, return_item=return_item):
                    yield item
        return []

    # returns (class_name, score, rect) from extra_data. Score and rect can be None!
    @staticmethod
    def enumerate_result_items(result, class_name_filters=None, use_wildcard=False, return_item=False) -> Iterator[
        tuple]:
        for item in result:
            class_name = item.get(constants.RESULT_KEY_CLASS_NAME, None)
            if class_name is not None:
                ok = False
                if class_name_filters is None:
                    ok = True
                else:
                    if use_wildcard:
                        for filter1 in class_name_filters:
                            if string_helper.wildcard(class_name, filter1):
                                ok = True
                                break
                    else:
                        ok = class_name in class_name_filters
                if ok:
                    score = item.get(constants.RESULT_KEY_SCORE, None)
                    rect = item.get(constants.RESULT_KEY_RECT, None)
                    if not return_item:
                        yield class_name, score, rect
                    else:
                        yield class_name, score, rect, item



import math
import cv2
import base64
import numpy as np

from ndu_gate_camera.utility import geometry_helper
from ndu_gate_camera.utility.ndu_utility import NDUUtility


def image_h_w(image):
    # h, w = image.shape[:2]
    return image.shape[:2]


# Boyut değişikliğine en uygun interpolation yöntemi ile resize eder.
def resize_best_quality(image, size):
    if image.shape[0] == size[0] and image.shape[1] == size[1]:
        return image.copy()
    size0 = max(image.shape[0], image.shape[1])
    size1 = max(size[0], size[1])
    if size0 > size1:
        # if size0 > 2 * size1:
        #     image = cv2.pyrDown(image)
        return cv2.resize(image, size, interpolation=cv2.INTER_AREA)
    else:
        return cv2.resize(image, size, interpolation=cv2.INTER_CUBIC)


# İmaj boyutlarından birisi max_dim'den daha büyükse küçültür, değilse aynen döner.
def resize_if_larger(image, max_dim, interpolation=None):
    h, w = image.shape[:2]
    if w > h:
        if w > max_dim:
            return resize(image, width=max_dim)
        else:
            return image
    else:
        if h > max_dim:
            return resize(image, height=max_dim)
        else:
            return image


# İmaj boyutlarından "büyük olan" min_dim'den daha küçükse resize edilir, değilse aynen döner.
def resize_if_smaller(image, min_dim, interpolation=None):
    h, w = image.shape[:2]
    if w > h:
        if w < min_dim:
            return resize(image, width=min_dim)
        else:
            return image
    else:
        if h < min_dim:
            return resize(image, height=min_dim)
        else:
            return image


# 'width' veya 'height yoksa en-boy oranını koruyarak resize eder. İkisi de varsa normal resize eder.
# 'interpolation' yoksa en uygununu seçer.
def resize(image, width=None, height=None, interpolation=None):
    if width is None and height is None:
        return image

    h, w = image.shape[:2]
    if width is None:
        r = height / float(h)
        dim = (int(w * r), height)
    elif height is None:
        r = width / float(w)
        dim = (width, int(h * r))
    else:
        dim = (width, height)

    if interpolation is None:
        return resize_best_quality(image, dim)
    else:
        return cv2.resize(image, dim, interpolation=interpolation)


# total_pixel_count sonucun width * height değeridir.
# int yuvarlama yüzünden sonuç w*h değer, tam total_pixel_count olmayabilir.
def resize_total_pixel_count(image, total_pixel_count):
    h, w = image.shape[:2]
    ratio = w / float(h)
    w1 = math.sqrt(total_pixel_count * ratio)
    h1 = w1 * h / float(w)
    return resize_best_quality(image, (int(w1), int(h1)))


# OpenCV mat nesnesini base64 string yapar
def to_base64(image):
    _, buffer = cv2.imencode('.jpg', image)
    return base64.b64encode(buffer)


# base64 string'i OpenCV mat nesnesi yapar
def from_base64(base64_text):
    original = base64.b64decode(base64_text)
    as_np = np.frombuffer(original, dtype=np.uint8)
    return cv2.imdecode(as_np, flags=1)


def fill_polyline_transparent(image, pnts, color, opacity, thickness=-1):
    blk = np.zeros(image.shape, np.uint8)
    cv2.drawContours(blk, pnts, -1, color, -1)
    if thickness >= 0:
        cv2.polylines(image, pnts, True, color=color, thickness=thickness)
    res = cv2.addWeighted(image, 1.0, blk, 0.1, 0)
    cv2.copyTo(res, None, image)


def select_areas(frame, window_name, color=(0, 0, 255), opacity=0.3, thickness=4, max_count=None, next_area_key="n", finish_key="s", return_tuples=True, max_point_count=None):
    try:
        areas = []
        area = []

        def get_mouse_points(event, x, y, _flags, _param):
            if event == cv2.EVENT_LBUTTONDOWN:
                area.append((x, y))

        cv2.namedWindow(window_name)
        cv2.moveWindow(window_name, 40, 30)
        cv2.setMouseCallback(window_name, get_mouse_points)

        new_area = False
        while True:
            image = frame.copy()
            for area1 in areas:
                pts = np.array(area1, np.int32)
                fill_polyline_transparent(image, [pts], color=color, opacity=opacity, thickness=thickness)

            if not new_area:
                if len(area) > 0:
                    pts = np.array(area, np.int32)
                    fill_polyline_transparent(image, [pts], color=color, opacity=opacity,
                                              thickness=thickness)
                    for pnt in area:
                        cv2.circle(image, pnt, thickness * 2, color, thickness)
            else:
                if len(area) > 2:
                    areas.append(area)
                if max_count is not None and len(areas) == max_count:
                    return areas
                else:
                    area = []
                    new_area = False

            cv2.imshow(window_name, image)
            k = cv2.waitKey(1)
            if k & 0xFF == ord(finish_key):
                break
            elif k & 0xFF == ord(next_area_key):
                new_area = True
            elif max_point_count is not None and len(area) == max_point_count:
                new_area = True

        if len(area) > 2:
            areas.append(area)
        if not return_tuples:
            for i in range(len(areas)):
                areas[i] = np.array(areas[i], np.int32)
        return areas
    finally:
        cv2.destroyWindow(window_name)


def select_lines(frame, window_name, color=(0, 255, 255), thickness=4, max_count=None, finish_key="s"):
    try:
        lines = []
        line = []

        def get_mouse_points(event, x, y, _flags, _param):
            if event == cv2.EVENT_LBUTTONDOWN:
                line.append((x, y))

        cv2.namedWindow(window_name)
        cv2.moveWindow(window_name, 40, 30)
        cv2.setMouseCallback(window_name, get_mouse_points)

        while True:
            image = frame.copy()
            for line1 in lines:
                pts = np.array(line1, np.int32)
                cv2.polylines(image, [pts], False, color=color, thickness=thickness)
            for pnt in line:
                cv2.circle(image, pnt, thickness * 2, color, thickness)
            if len(line) == 2:
                lines.append(line)
                if max_count is not None and len(lines) == max_count:
                    return lines
                else:
                    line = []

            cv2.imshow(window_name, image)
            k = cv2.waitKey(1)
            if k & 0xFF == ord(finish_key):
                break

        return lines
    finally:
        cv2.destroyWindow(window_name)


def select_points(frame, window_name, color=(0, 255, 255), radius=8, thickness=4, max_count=None, finish_key="s"):
    try:
        pnts = []

        def get_mouse_points(event, x, y, _flags, _param):
            if event == cv2.EVENT_LBUTTONDOWN:
                pnts.append((x, y))

        cv2.namedWindow(window_name)
        cv2.moveWindow(window_name, 40, 30)
        cv2.setMouseCallback(window_name, get_mouse_points)

        while True:
            image = frame.copy()
            for pnt in pnts:
                cv2.circle(image, pnt, radius, color, thickness)
            cv2.imshow(window_name, image)
            k = cv2.waitKey(1)
            if k & 0xFF == ord(finish_key):
                break
            if max_count is not None and max_count <= len(pnts):
                break
        return pnts
    finally:
        cv2.destroyWindow(window_name)


def put_text(img, text_, center, color=None, font_scale=0.5, thickness=1, back_color=None, replace_tur_chars=True):
    if replace_tur_chars:
        text_ = NDUUtility.debug_replace_tur_chars(text_)
    if back_color is None:
        back_color = [1, 1, 1]
    if color is None:
        color = [255, 255, 255]
    y = center[1]
    # font = cv2.FONT_HERSHEY_COMPLEX
    font = cv2.FONT_HERSHEY_DUPLEX
    coor = (int(center[0] + 5), int(y))
    cv2.putText(img=img, text=text_, org=coor,
                fontFace=font, fontScale=font_scale, color=back_color, lineType=cv2.LINE_AA,
                thickness=thickness + 2)
    cv2.putText(img=img, text=text_, org=coor,
                fontFace=font, fontScale=font_scale, color=color,
                lineType=cv2.LINE_AA, thickness=thickness)


def rescale_frame(frame, percent):
    width = int(frame.shape[1] * percent / 100.0)
    height = int(frame.shape[0] * percent / 100.0)
    dim = (width, height)
    return resize_best_quality(frame, dim)


def frame2base64(frame, scale=40):
    scaled_frame = rescale_frame(frame, scale)
    res, frame = cv2.imencode('.png', scaled_frame)
    base64_data = base64.b64encode(frame)
    return base64_data.decode('utf-8')


def normalize_coordinates(image, coords):
    h, w = image.shape[:2]
    pnts = []
    for x, y in coords:
        pnts.append((x / w, y / h))
    return pnts


def denormalize_coordinates(image, coords):
    h, w = image.shape[:2]
    pnts = []
    for x, y in coords:
        pnts.append((int(x * w), int(y * h)))
    return pnts


def normalize_lines(image, lines):
    lines1 = []
    for line in lines:
        lines1.append(normalize_coordinates(image, line))
    return lines1


def denormalize_lines(image, lines):
    lines1 = []
    for line in lines:
        lines1.append(denormalize_coordinates(image, line))
    return lines1


def convert_lines_list2tuple(lines):
    res = []
    for line in lines:
        line1 = []
        for c in line:
            line1.append(tuple(c))
        res.append(line1)
    return res


def convert_lines_tuple2list(lines):
    res = []
    for line in lines:
        line1 = []
        for c in line:
            line1.append(list(c))
        res.append(line1)
    return res


def crop(frame, rect):
    y1 = max(int(rect[0]), 0)
    x1 = max(int(rect[1]), 0)
    y2 = max(int(rect[2]), 0)
    x2 = max(int(rect[3]), 0)
    return frame[y1:y2, x1:x2]


def change_brightness(img, value):
    hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(hsv)

    if value > 0:
        lim = 255 - value
        v[v > lim] = 255
        v[v <= lim] += value
    else:
        lim = 0 - value
        v[v < lim] = 0
        v[v >= lim] -= abs(value)

    final_hsv = cv2.merge((h, s, v))
    img = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2BGR)
    return img


def get_mask(shape, areas):
    h, w = shape[:2]
    mask = np.zeros((h, w), np.uint8)
    contours = []
    for area in areas:
        contours.append(np.array(area, np.int32))
    cv2.drawContours(mask, contours, -1, 255, -1)
    return mask


def apply_mask(image, mask):
    return cv2.bitwise_and(image, image, mask=mask)


def draw_rect(image, rect, color=None):
    if color is None:
        color = [255, 255, 255]
    c = np.array(rect[:4], dtype=np.int32)
    c1, c2 = geometry_helper.get_rect_pnts(c)
    cv2.rectangle(image, c1, c2, color=[1, 1, 1], thickness=3)
    cv2.rectangle(image, c1, c2, color=color, thickness=2)


def equalize_hist_bgr(img):
    img_yuv = cv2.cvtColor(img, cv2.COLOR_BGR2YUV)
    img_yuv[:, :, 0] = cv2.equalizeHist(img_yuv[:, :, 0])
    return cv2.cvtColor(img_yuv, cv2.COLOR_YUV2BGR)

    # img[:, :, 0] = cv2.equalizeHist(img[:, :, 0])
    # img[:, :, 1] = cv2.equalizeHist(img[:, :, 1])
    # img[:, :, 2] = cv2.equalizeHist(img[:, :, 2])
    # return img


def adaptive_equalize_hist_bgr(img):
    clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
    img_yuv = cv2.cvtColor(img, cv2.COLOR_BGR2YUV)
    img_yuv[:, :, 0] = clahe.apply(img_yuv[:, :, 0])
    return cv2.cvtColor(img_yuv, cv2.COLOR_YUV2BGR)

    # clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
    # img[:, :, 0] = clahe.apply(img[:, :, 0])
    # img[:, :, 1] = clahe.apply(img[:, :, 1])
    # img[:, :, 2] = clahe.apply(img[:, :, 2])
    # return img



from os import path
import cv2
import vidgear
import time

from ndu_gate_camera.utility.ndu_utility import NDUUtility

try:
    from vidgear.gears import CamGear
except ImportError:
    if NDUUtility.install_package("vidgear") == 0:
        from vidgear.gears import CamGear

from ndu_gate_camera.api.video_source import VideoSource, log


class YoutubeVideoSource(VideoSource):
    def __init__(self, source_config):
        """

        :param source_config:
        """
        super().__init__()
        log.info("source_config %s", source_config)
        self.__show_preview = source_config.get("show_preview", False)
        self.__url = source_config.get("url", None)
        if self.__url is None:
            raise ValueError("Video url is empty")

        if not NDUUtility.is_url_valid(url=self.__url):
            raise ValueError("Video url is not valid:  %s0", self.__url)

        self.__cam_gear_options = source_config.get("cam_gear_options", {"CAP_PROP_FRAME_WIDTH ": 320, "CAP_PROP_FRAME_HEIGHT": 240, "CAP_PROP_FPS ": 1})

        self.__stream = None
        self._set_capture()

    def get_frames(self):
        log.debug("start video streaming..")
        count = 0
        # TODO - bitince başa sar?
        self.__stream.start()
        try:
            while True:
                frame = self.__stream.read()
                if frame is None:
                    break
                yield count, frame
                count += 1
        except StopIteration:
            pass
        finally:
            self.stop()

    def reset(self):
        self.__stream.stop()
        self._set_capture()

    def stop(self):
        log.debug("Youtube video finished..")
        if self.__show_preview:
            cv2.destroyAllWindows()
        self.__stream.stop()

    def _set_capture(self):
        self.__stream = CamGear(source=self.__url, y_tube=True, time_delay=1, logging=True, **self.__cam_gear_options)



from ndu_gate_camera.api.result_handler import ResultHandler, log
from os import path


class ResultHandlerFile(ResultHandler):
    def __init__(self, folder):
        self.__working_path = folder
        if not path.isdir(self.__working_path):
            log.warning("This path is not a folder %s", self.__working_path)
            self.__working_path = "/etc/ndu_gate/"

        self.__telemetry_file = self.__working_path + 'serviceTelemetry.txt'

        self.__working_path = path.expanduser("~") + path.sep + 'serviceTelemetry.txt'

        if not path.isfile(self.__telemetry_file):
            try:
                file = open(self.__telemetry_file, 'w+')
                file.close()
            except Exception as e:
                log.error(e)
                log.warning("Can not create telemetry file %s", self.__telemetry_file)

        log.info("ResultHandlerFile %s", self.__telemetry_file)

    def save_result(self, result, device=None, runner_name=None, data_type='telemetry'):
        """
        Verilen ölçümleri serviceTelemetry.txt dosyasına yazar.
        :param self:
        :param result: {"key" : "telem-key", "value": 1, "ts" : timestamp }
        :param device:
        :param runner_name:
        :param data_type:
        :return:
        """
        with open(self.__telemetry_file, 'r+') as f:
            f.seek(0)
            for i in range(len(result)):
                f.write(str(result[i]) + '\n')
            f.truncate()
            f.close()

    def clear_results(self, runner_name=None):
        pass



import numpy as np
import cv2

from ndu_gate_camera.api.video_source import VideoSource, log
from ndu_gate_camera.utility.ndu_utility import NDUUtility

# try:
#     from picamera import PiCamera
# except ImportError:
#     print("picamera library not found - installing...")
#     if NDUUtility.install_package("picamera"):
#         import picamera
#
# try:
#     from picamera.array import PiRGBArray
# except ImportError:
#     print("picamera library not found - installing...")
#     if NDUUtility.install_package("picamera[array]"):
#         import picamera.array


class PiCameraVideoSource(VideoSource):
    def __init__(self, frame_width=640, frame_height=480, framerate=32, rotation=180, show_preview=False):
        super().__init__()
        self.__camera = None
        self.__frame_width = frame_width
        self.__frame_height = frame_height
        self.__framerate = framerate
        self.__rotation = rotation
        self.__show_preview = show_preview

        self._set_capture()

    def get_frames(self):
        log.info("start camera streaming..")
        frame_num = 0

        for frameFromCam in self.__camera.capture_continuous(self.__rawCapture, format="bgr", use_video_port=True):
            try:
                frame = np.copy(frameFromCam.array)
                frame_num += 1
                frame_h = frame.shape[0]
                frame_w = frame.shape[1]
                yield frame_num, frame
            except KeyboardInterrupt:
                self.__rawCapture.truncate(0)
                self.__camera.close()
                cv2.destroyAllWindows()
                log.info("exit from pi camera streaming")
                break

            log.info("finish pi camera streaming")

    def reset(self):
        pass

    def stop(self):
        pass

    def _set_capture(self):
        self.__camera = PiCamera()
        self.__camera.resolution = (self.__frame_width, self.__frame_height)
        self.__camera.framerate = self.__framerate
        self.__camera.rotation = self.__rotation
        self.__rawCapture = PiRGBArray(self.__camera, size=(self.__frame_width, self.__frame_height))



from threading import Lock

locks_lock = Lock()
locks = {}


def get_lock(value):
    global locks
    with locks_lock:
        if value not in locks:
            locks[value] = Lock()
        return locks[value]



import json
import time
import zmq

from ndu_gate_camera.api.result_handler import log
from ndu_gate_camera.utility import constants


class PlatformHandlerSocket:
    def __init__(self, config, device_name):
        self.__host = config.get("host", "https://ndu-test.netcad.com")

        log.info("PlatformHandlerSocket %s ", self.host)

    def save_result(self, result, device=None, runner_name=None, data_type='telemetry'):
        """

        :param self:
        :param result: {"key" : "telem-key", "value": 1, "ts" : timestamp }
        :param device:
        :param runner_name:
        :param data_type: telemetry or attribute
        :return:
        """
        if device is None:
            device = self.__device_name
        try:
            if result is not None:
                for item in result:
                    data = item.get(constants.RESULT_KEY_DATA, None)
                    if data is not None:
                        self.__send_item(data, device, runner_name, data_type)

        except Exception as e:
            log.error(e)

    def __send_item(self, item, device, runner_name, data_type):
        data = {
            data_type: item,
            "runner": runner_name,
            "deviceName": device
        }
        data_json = json.dumps(data)
        try:
            self.socket.send_string("ndugate " + data_json)
        except Exception as a:
            log.debug("exception sending item : %s", a)

    def clear_results(self, runner_name=None):
        pass



import threading
import time

import cv2

from ndu_gate_camera.api.video_source import VideoSource, log


class IPCameraVideoSource(VideoSource):
    def __init__(self, source_config):
        super().__init__()
        self.__video_url = source_config.get("url", None)
        if self.__video_url is None:
            raise ValueError("Video url is empty")
        self._set_capture()
        self.mode = source_config.get("mode", 1)  # 1 - QUEUE , 0 = SEND ALL
        self.count = 0
        self.frame = None
        if self.mode != 0:
            p1 = threading.Thread(target=self.read_frame)
            p1.start()

    def read_frame(self):
        while self.__capture.isOpened():
            ret, frame = self.__capture.read()
            if ret is False:
                break

            self.frame = frame
            self.count += 1

        self.count = -1
        self.__capture.release()

    def get_frames(self):
        log.debug("start ip camera video streaming..")

        if self.mode == 0:
            while self.__capture.isOpened():
                ret, frame = self.__capture.read()
                if ret is False:
                    break

                yield self.count, frame
                self.count += 1
        else:
            while self.count == 0:
                time.sleep(1)
            if self.count == -1:
                raise Exception('Cannot capture ip camera')
            while True:
                yield self.count, self.frame

        self.__capture.release()

    def reset(self):
        self._set_capture()

    def stop(self):
        # TODO
        pass

    def _set_capture(self):
        # TODO - https://www.pyimagesearch.com/2019/04/15/live-video-streaming-over-network-with-opencv-and-imagezmq/
        try:
            self.__capture = cv2.VideoCapture(self.__video_url)
        except Exception as exp:
            log.error(exp)




from sys import stdout
import logging
import logging.handlers


class NDULoggerHandler(logging.Handler):
    def __init__(self, gateway):
        self.current_log_level = 'INFO'
        super().__init__(logging.getLevelName(self.current_log_level))
        self.setLevel(logging.getLevelName('DEBUG'))
        self.__gateway = gateway
        self.activated = False
        self.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - [%(filename)s] - %(module)s - %(lineno)d - %(message)s'))
        self.loggers = ['service',
                        'runner',
                        'connector',
                        'video_source'
                        ]
        for logger in self.loggers:
            log = logging.getLogger(logger)
            log.addHandler(self.__gateway.main_handler)
            log.debug("Added remote handler to log %s", logger)

    def activate(self, log_level=None):
        try:
            for logger in self.loggers:
                if log_level is not None and logging.getLevelName(log_level) is not None:
                    log = logging.getLogger(logger)
                    self.current_log_level = log_level
                    log.setLevel(logging.getLevelName(log_level))
        except Exception as e:
            log = logging.getLogger('service')
            log.exception(e)
        self.activated = True

    def handle(self, record):
        if self.activated:
            record = self.formatter.format(record)

    def deactivate(self):
        self.activated = False

    @staticmethod
    def set_default_handler():
        logger_names = [
            'service',
            'runner',
            'connector',
            'video_source'
        ]
        for logger_name in logger_names:
            logger = logging.getLogger(logger_name)
            handler = logging.StreamHandler(stdout)
            handler.setFormatter(logging.Formatter('[STREAM ONLY] %(asctime)s - %(levelname)s - [%(filename)s] - %(module)s - %(lineno)d - %(message)s'))
            logger.addHandler(handler)



import getopt
import logging
import os
import sys
import time
import traceback
from os import path, listdir, mkdir, curdir
from sys import platform

from yaml import safe_load

from ndu_gate_camera.camera.ndu_camera_service import NDUCameraService
from ndu_gate_camera.camera.ndu_logger import NDULoggerHandler
from ndu_gate_camera.camera.result_handlers.result_handler_file import ResultHandlerFile
from ndu_gate_camera.camera.result_handlers.result_handler_socket import ResultHandlerSocket
from ndu_gate_camera.camera.result_handlers.result_handler_mqtt import ResultHandlerMqtt
from ndu_gate_camera.utility import onnx_helper
from ndu_gate_camera.utility.constants import DEFAULT_NDU_GATE_CONF, DEFAULT_NDU_GATE_CONF_WIN, DEFAULT_HANDLER_SETTINGS
from ndu_gate_camera.utility.ndu_utility import NDUUtility


def _get_config():
    print(sys.argv)
    ndu_gate_config_file = None

    if len(sys.argv) > 1:
        try:
            opts, _ = getopt.getopt(sys.argv[1:], "c:", ["config="])
            for opt, arg in opts:
                if opt in ['-c', '--config']:
                    ndu_gate_config_file = arg
        except getopt.GetoptError:
            print('ndu_camera.py -c <config_file_path>')

    if ndu_gate_config_file and path.sep not in ndu_gate_config_file:
        ndu_gate_config_file = path.join(path.dirname(path.abspath(__file__)) + '/config/'.replace('/', path.sep),
                                         ndu_gate_config_file)

    config_file_name = "ndu_gate.yaml"
    if ndu_gate_config_file is None:
        config_file = path.dirname(path.abspath(__file__)) + '/config/'.replace('/', path.sep) + config_file_name
        if path.isfile(config_file):
            ndu_gate_config_file = config_file

    if ndu_gate_config_file is None:
        print("Config file not specified, going to use default")
        if platform == "win32":
            ndu_gate_config_file = DEFAULT_NDU_GATE_CONF_WIN.replace('/', path.sep)
        else:
            ndu_gate_config_file = DEFAULT_NDU_GATE_CONF.replace('/', path.sep)

    print("Config file is {}".format(ndu_gate_config_file))

    return ndu_gate_config_file


def main(ndu_gate_config_file):
    if "logs" not in listdir(curdir):
        mkdir("logs")

    try:
        if ndu_gate_config_file is None:
            ndu_gate_config_file = path.dirname(path.dirname(path.abspath(__file__))) + '/config/ndu_gate.yaml'.replace(
                '/', path.sep)

        if not path.isfile(ndu_gate_config_file):
            print('config parameter is not a file : ', ndu_gate_config_file)
            sys.exit(2)

        print("Using config file : {}".format(ndu_gate_config_file))
        with open(ndu_gate_config_file, encoding="utf-8") as general_config:
            ndu_gate_config = safe_load(general_config)

        ndu_gate_config_dir = path.dirname(path.abspath(ndu_gate_config_file)) + path.sep

        logging_config_file = ndu_gate_config_dir + "logs.conf"
        if NDUUtility.is_debug_mode():
            logging_config_file = ndu_gate_config_dir + "logs_debug.conf"
        try:
            import platform
            if platform.system() == "Darwin":
                ndu_gate_config_dir + "logs_macosx.conf"
            # logging.config.fileConfig(logging_config_file, disable_existing_loggers=False)
        except Exception as e:
            print(e)
            NDULoggerHandler.set_default_handler()

        global log
        log = logging.getLogger('service')
        log.info("NDUCameraService starting...")
        log.info("NDU-Gate logging config file: %s", logging_config_file)
        log.info("NDU-Gate logging service level: %s", log.level)

        onnx_helper.init(ndu_gate_config.get("onnx_runner", None), log)
        extension_folder = ndu_gate_config.get("extension_folder", None)

        result_hand_conf = ndu_gate_config.get("result_handler", None)
        if result_hand_conf is None:
            result_hand_conf = DEFAULT_HANDLER_SETTINGS

        res_handler = result_hand_conf.get("type")

        if str(res_handler) == str("SOCKET"):
            result_handler = ResultHandlerSocket(result_hand_conf.get("socket", {}),
                                                 result_hand_conf.get("device", None))
        elif res_handler == str("MQTT"):
            print('RESULT HANDLER IS MQTT!!!')
            result_handler = ResultHandlerMqtt(result_hand_conf.get("access_token"), result_hand_conf.get("mqtt"))
        elif str(res_handler) == str("FILE"):
            result_handler = ResultHandlerFile(result_hand_conf.get("file_path", None))
        else:
            result_handler = ResultHandlerFile(result_hand_conf.get("file_path", None))

        instances = ndu_gate_config.get("instances")
        if len(instances) > 1:
            services = []
            preview_exists = False
            for instance in instances:
                if instance["source"].get("ignore", True):
                    continue
                if instance["source"].get("preview_show", False):
                    preview_exists = True
                camera_service = NDUCameraService(instance=instance, config_dir=ndu_gate_config_dir,
                                                  handler=result_handler, is_main_thread=False,
                                                  extension_folder=extension_folder)
                camera_service.start()
                services.append(camera_service)
                log.info("NDU-Gate an instance started")

            log.info("NDU-Gate all instances are started")
            if preview_exists:
                alive_exists = True
                while alive_exists:
                    time.sleep(0.033333)
                    alive_exists = False
                    for s in services:
                        if s.is_alive():
                            alive_exists = True
                            s.check_for_preview()
                        else:
                            s.finish_preview()
            else:
                for service in services:
                    service.join()
        elif len(instances) == 1 and instances[0]["source"].get("ignore", True) is False:
            camera_service = NDUCameraService(instance=instances[0], config_dir=ndu_gate_config_dir,
                                              handler=result_handler, is_main_thread=True,
                                              extension_folder=extension_folder)
            camera_service.run()
        else:
            log.error("NDUCameraService no source found!")

        log.info("NDUCameraService exiting...")

    except Exception as e:
        print("NDUCameraService Exited")
        print(e)
        print("----------------------")
        print(traceback.format_exc())


def daemon():
    print("Starting as daemon")
    main(_get_config())


def daemon_with_conf(config_file):
    print("Starting as daemon with conf {} ".format(config_file))
    main(config_file)


if __name__ == '__main__':
    main(_get_config())



from abc import ABC, abstractmethod

import logging

log = logging.getLogger("runner")


class NDUCameraRunner(ABC):
    frame_count = 0

    def __init__(self):
        self.frame_count = 0

    @abstractmethod
    def get_name(self):
        pass

    @abstractmethod
    def get_settings(self):
        pass

    @abstractmethod
    def process_frame(self, frame, extra_data=None):
        """
        :param frame: tek bir frame datası
        :param extra_data pre process sırasında frameden çıkarılan veriler, person, face gibi key ler içeren dictionary
        """
        self.frame_count = self.frame_count + 1
        log.debug("%s processing frame %s", self.get_name(), self.frame_count)



from abc import ABC, abstractmethod

import logging
from enum import Enum

log = logging.getLogger("video_source")


class VideoSource(ABC):

    @abstractmethod
    def get_frames(self):
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def stop(self):
        pass


class VideoSourceType(Enum):
    VIDEO_FILE = 1
    PI_CAMERA = 2
    CAMERA = 3
    IP_CAMERA = 4
    VIDEO_URL = 5
    YOUTUBE = 6
    IMAGE_FILE = 7



class FramePreProcessors:
    """
    Herbir frame için yapılacak olan ön işleme metodları burada bulunur.
    """
    def __init__(self):
        pass

    @staticmethod
    def find_persons(self, frame):
        pass

    @staticmethod
    def find_faces(self, frame):
        pass



import math
import cv2
import numpy as np


def is_inside_polygon(polygon, point):
    # ref:https://stackoverflow.com/a/2922778/1266873
    # int pnpoly(int nvert, float *vertx, float *verty, float testx, float testy)
    # {
    #   int i, j, c = 0;
    #   for (i = 0, j = nvert-1; i < nvert; j = i++) {
    #     if ( ((verty[i]>testy) != (verty[j]>testy)) &&
    #      (testx < (vertx[j]-vertx[i]) * (testy-verty[i]) / (verty[j]-verty[i]) + vertx[i]) )
    #        c = !c;
    #   }
    #   return c;
    # }
    def pnpoly(nvert, vertx, verty, testx, testy):
        i = 0
        j = nvert - 1
        c = False
        while True:
            j = i
            i += 1
            if i >= nvert:
                break
            if (verty[i] > testy) != (verty[j] > testy) and testx < (vertx[j] - vertx[i]) * (testy - verty[i]) / (
                    verty[j] - verty[i]) + vertx[i]:
                c = not c
        return c

    vertx = []
    verty = []
    for p in polygon:
        vertx.append(p[0])
        verty.append(p[1])
    if polygon[-1] is not polygon[0]:
        p = polygon[0]
        vertx.append(p[0])
        verty.append(p[1])
    return pnpoly(len(vertx), vertx, verty, point[0], point[1])


def is_inside_rect(rect, point):
    e = 0.001
    x = point[0]
    if x < rect[1] - e:
        return False
    elif x > rect[3] + e:
        return False
    y = point[1]
    if y < rect[0] - e:
        return False
    elif y > rect[2] + e:
        return False
    return True


def rects_intersect(rect1, rect2):
    class Rectangle:
        def intersects(self, other):
            a, b = self, other
            x1 = max(min(a.x1, a.x2), min(b.x1, b.x2))
            y1 = max(min(a.y1, a.y2), min(b.y1, b.y2))
            x2 = min(max(a.x1, a.x2), max(b.x1, b.x2))
            y2 = min(max(a.y1, a.y2), max(b.y1, b.y2))
            return x1 < x2 and y1 < y2

        def _set(self, x1, y1, x2, y2):
            if x1 > x2 or y1 > y2:
                raise ValueError("Coordinates are invalid")
            self.x1, self.y1, self.x2, self.y2 = x1, y1, x2, y2

        def __init__(self, bbox):
            self._set(bbox[0], bbox[1], bbox[2], bbox[3])

    return Rectangle(rect1).intersects(Rectangle(rect2))


def rects_overlap(rect1, rect2, min_ratio):
    w1 = rect1[2] - rect1[0]
    w2 = rect2[2] - rect2[0]
    w_min = min(rect1[2], rect2[2]) - max(rect1[0], rect2[0])
    if (w_min / w1) >= min_ratio or (w_min / w2) >= min_ratio:
        h1 = rect1[3] - rect1[1]
        h2 = rect2[3] - rect2[1]
        h_min = min(rect1[3], rect2[3]) - max(rect1[1], rect2[1])
        return (h_min / h1) >= min_ratio or (h_min / h2) >= min_ratio
    return False


def add_padding_rect(rect, padding):
    x1, y1, x2, y2 = rect[0], rect[1], rect[2], rect[3]
    dw = (x2 - x1) * padding
    dh = (y2 - y1) * padding
    return [x1 - dw, y1 - dh, x2 + dw, y2 + dh]


def distance(p1, p2):
    return math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)


def get_rect_values(rect):
    # y1, x1, y2, x2
    return rect[0], rect[1], rect[2], rect[3]


def get_rect_pnts(rect):
    # p1, p2
    return (rect[1], rect[0]), (rect[3], rect[2])


def get_rect(p1, p2):
    return [p1[1], p1[0], p2[1], p2[0]]


def get_rect_bottom_center(rect):
    r1 = rect[1]
    r2 = rect[2]
    r3 = rect[3]
    return int(r1 + (r3 - r1) * 0.5), r2  # bottom center


def get_dist_sq(p1, p2):
    d1 = p2[0] - p1[0]
    d2 = p2[1] - p1[1]
    return d1 * d1 + d2 * d2


def get_center_int(pnts):
    x = 0
    y = 0
    for pnt in pnts:
        x += pnt[0]
        y += pnt[1]
    length = float(len(pnts))
    return [int(x / length), int(y / length)]



import sys
from threading import Thread

import cv2
import numpy as np

from ndu_gate_camera.api.ndu_camera_runner import NDUCameraRunner
from ndu_gate_camera.utility import constants, geometry_helper, image_helper
from ndu_gate_camera.utility.ndu_utility import NDUUtility


class ROIManager():
    def __init__(self, config, name):
        self.Name = name
        self._polygons = config.get("polygons", [])
        self._apply_mask = config.get("apply_mask", True)
        self._apply_crop = config.get("apply_crop", True)
        self._apply_histogram_equalization = config.get("apply_histogram_equalization", False)
        self._apply_adaptive_histogram_equalization = config.get("apply_adaptive_histogram_equalization", False)
        self._preview = config.get("preview", False)
        self._pyrUp = config.get("pyrUp", 0)
        self._select_polygons_mode = config.get("select_polygons_mode", False)

        self._inited = False
        self._mask = None
        self._roi = None
        self._last_frame = None
        self._rev_div = 1

    def forward(self, frame):
        self._last_frame = frame
        if not self._inited:
            self._inited = True
            if self._select_polygons_mode:
                polygons = image_helper.select_areas(frame, "Select points for roi config {}".format(self.Name), max_count=None, next_area_key="n", finish_key="s", return_tuples=False)
                arr1 = []
                for polygon in polygons:
                    arr = []
                    for p in polygon:
                        arr.append(list(p))
                    arr1.append(arr)
                print('"polygons": {},'.format(str(arr1)))
                exit(1)

            min_max = []
            for i in range(len(self._polygons)):
                area = np.array(self._polygons[i], np.int32)
                self._polygons[i] = area
                min_max.append(area.min(axis=0, initial=None))
                min_max.append(area.max(axis=0, initial=None))
            min_max = np.array(min_max, np.int32)
            min_ = min_max.min(axis=0, initial=None)
            max_ = min_max.max(axis=0, initial=None)

            if self._apply_crop:
                self._roi = [min_[1], min_[0], max_[1], max_[0]]
            if self._apply_mask:
                self._mask = image_helper.get_mask(frame.shape, self._polygons)
                if self._roi is not None:
                    self._mask = image_helper.crop(self._mask, self._roi)

        # masked = image_helper.apply_mask(frame, self._mask)
        img = frame
        if self._apply_crop:
            img = image_helper.crop(img, self._roi)
        if self._apply_mask:
            img = image_helper.apply_mask(img, self._mask)
        if self._pyrUp > 0:
            self._rev_div = 1
            for i in range(self._pyrUp):
                self._rev_div *= 2
                img = cv2.pyrUp(img, img)

        if self._apply_histogram_equalization:
            img = image_helper.equalize_hist_bgr(img)
        if self._apply_adaptive_histogram_equalization:
            img = image_helper.adaptive_equalize_hist_bgr(img)

        if self._preview:
            cv2.imshow("ROIManager", img)
        return img

    def reverse(self, frame, results):
        if self._apply_crop and results is not None and len(results) > 0:
            for result in results:
                if "rect" in result:
                    rect = result["rect"]
                    if self._rev_div > 1:
                        for i in range(len(rect)):
                            rect[i] = int(rect[i] / self._rev_div)
                    rect[0] += self._roi[0]
                    rect[1] += self._roi[1]
                    rect[2] += self._roi[0]
                    rect[3] += self._roi[1]
        return self._last_frame, results



import time

# import paho.mqtt.client as mqtt
import json

import paho.mqtt.client as mqttClient
from ndu_gate_camera.utility import constants


class ResultHandlerMqtt:
    '''
        Cihaz verilerinin NDU platformuna MQTT APİ üzerinden gönderilmesi
    '''

    def __init__(self, access_token, mqtt_obj):
        host = mqtt_obj.get("host")

        self.connected_devices = {}

        self.client = mqttClient.Client()
        self.client.username_pw_set(access_token)
        # Connect to ThingsBoard using default MQTT port and 60 seconds keepalive interval
        try:
            self.client.connect(host, mqtt_obj.get("port"), 60)
            self.client.loop_start()
        except:
            print("can not connect mqtt broker!")
            # TODO - bağlanamasa belli aralıklara bağlanmayı tekrar denesin, bağlanamadığ sırada veri gelirse o veriyi hafızada belirli bir süre/miktar tutmalı.
            #  bağlantı oluşur oluşmaz bu biriken veriler de gönderilmelidir.


    def send_connect_request(self, device_name):
        mqtt_data = {"device": device_name}
        self.client.publish('v1/gateway/connect', json.dumps(mqtt_data), 1)
        self.connected_devices[device_name] = True

    def send_disconnect_request(self, device_name):
        mqtt_data = {"device": device_name}
        self.client.publish('v1/gateway/disconnect', json.dumps(mqtt_data), 1)
        self.connected_devices.pop(device_name)

    def save_result(self, results, device=None, runner_name=None, data_type='telemetry'):
        if self.connected_devices.get(device) is None:
            self.send_connect_request(device)

        if data_type == 'telemetry':
            self.send_telemetry(device, results)
        elif data_type == 'attribute':
            self.send_attribute(device, results)

    def send_telemetry(self, device, results):
        try:
            mqtt_data = {
                device: []
            }
            # if mqtt_data.get(device) is None:
            #     mqtt_data['device'] = 'device'

            for result in results:
                if result is None:
                    continue
                data = result.get(constants.RESULT_KEY_DATA, None)

                if data is None:
                    continue

                single_data = {}
                for key in data:
                    if data[key] is not None:
                        single_data[key] = data[key]

                mqtt_data[device].append(single_data)
            if not len(mqtt_data[device]) == 0:
                self.client.publish('v1/gateway/telemetry', json.dumps(mqtt_data), 1)
        except KeyError:
            print('Exception while saving result, key error')

    def send_attribute(self, device, results):
        try:
            mqtt_data = {
                device: {}
            }

            for result in results:
                if result is None:
                    continue
                data = result.get(constants.RESULT_KEY_DATA, None)

                if data is None:
                    continue

                for key in data:
                    if data[key] is not None:
                        mqtt_data[device][key] = data[key]

            if not len(mqtt_data[device]) == 0:
                self.client.publish('v1/gateway/attributes', json.dumps(mqtt_data), 1)
        except KeyError:
            print('Exception while saving result, key error')

    def dispose(self):
        # TODO - disconnect devices
        pass



import time
import urllib.request
import os
from os import path
import zipfile


def download_runner_data(runner_type, runner_load_path, download_urls):
    name = runner_type

    runner_dir = runner_load_path
    if not runner_load_path:
        runner_dir = path.join(get_runner_dir(), name)

    if download_urls.get(runner_type) is None:
        return

    url = download_urls[name]

    data_folder = path.join(runner_dir, 'data')
    zip_file_path = os.path.join(runner_dir, 'data.zip')

    # if os.path.isfile(zip_file_path):
    #     old_zip_file_path = os.path.join(runner_dir, 'old_' + time.time() + '_data.zip')
    #     os.rename(zip_file_path, old_zip_file_path)

    try:
        if path.isdir(data_folder) is False:
            print('Data folder created')
            os.mkdir(data_folder)

        if not path.isfile(zip_file_path):
            print('Path to download is : ', data_folder)
            urllib.request.urlretrieve(url, zip_file_path)
        else:
            print('Data zip file exists, skipping download')

        if len(os.listdir(data_folder)) > 0:
            print("Data folder is not empty")
        else:
            unzip_to_data_folder(zip_file_path, runner_dir)

    except urllib.error.HTTPError as e:
        print(e.code)
    except PermissionError as e:
        print('PermissionError during extracting zip file')
    except Exception as e:
        print('Error during retrieving file')
        print(e)




def get_runner_dir():
    cur_dir = os.getcwd()
    parent_dir = os.path.abspath(os.path.join(cur_dir, os.pardir))
    runner_dir = os.path.join(parent_dir, 'runners')
    return runner_dir


def unzip_to_data_folder(zip_file_path, data_folder):
    zip_ref = zipfile.ZipFile(zip_file_path, 'r')
    zip_ref.extractall(data_folder)
    zip_ref.close()

# def unzip_and_delete(runner_dir, name):
#     # unzip
#     if path.isdir(os.path.join(runner_dir, name)) is False or path.isdir(
#             os.path.join(runner_dir, name, 'data')) is False:
#         zip_ref = zipfile.ZipFile(os.path.join(runner_dir, name + '.zip'), 'r')
#         zip_ref.extractall(os.path.join(runner_dir))
#         zip_ref.close()
#
#         # delete zip
#         os.remove(os.path.join(runner_dir, name + '.zip'))



import uuid

import numpy as np
import cv2
import time

from ndu_gate_camera.utility import constants, onnx_helper, image_helper, file_helper


def predict_v5(sess_tuple, input_size, class_names, frame):
    # https://github.com/ultralytics/yolov5
    # torch ve torchvision bağımlılığını kaldırmak için birçok değişiklik yapılmıştır, kod üstteki linktekinden değişiktir.

    def image_preprocess(image1, target_size):
        ih, iw = target_size
        h1, w1, _ = image1.shape

        scale = min(iw / w1, ih / h1)
        nw, nh = int(scale * w1), int(scale * h1)
        if nh != h1 or nw != w1:
            image_resized = image_helper.resize_best_quality(image1, (nw, nh))
        else:
            image_resized = image1

        image_padded = np.full(shape=[ih, iw, 3], fill_value=128.0)
        dw, dh = (iw - nw) // 2, (ih - nh) // 2
        image_padded[dh:nh + dh, dw:nw + dw, :] = image_resized
        image_padded = image_padded / 255.

        return image_padded, w1, h1

    def non_max_suppression(prediction, conf_thres=0.1, iou_thres=0.6, agnostic=False):
        """Performs Non-Maximum Suppression (NMS) on inference results

        Returns:
             detections with shape: nx6 (x1, y1, x2, y2, conf, cls)
        """

        def xywh2xyxy(x_):
            # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
            y = np.zeros_like(x_)
            y[:, 0] = x_[:, 0] - x_[:, 2] / 2  # top left x
            y[:, 1] = x_[:, 1] - x_[:, 3] / 2  # top left y
            y[:, 2] = x_[:, 0] + x_[:, 2] / 2  # bottom right x
            y[:, 3] = x_[:, 1] + x_[:, 3] / 2  # bottom right y
            return y

        def nms_cpu(boxes_, confs_, nms_thresh=0.5, min_mode=False):
            # print(boxes.shape)
            x1_ = boxes_[:, 0]
            y1_ = boxes_[:, 1]
            x2_ = boxes_[:, 2]
            y2_ = boxes_[:, 3]

            areas = (x2_ - x1_) * (y2_ - y1_)
            order = confs_.argsort()[::-1]

            keep = []
            while order.size > 0:
                idx_self = order[0]
                idx_other = order[1:]

                keep.append(idx_self)

                xx1 = np.maximum(x1_[idx_self], x1_[idx_other])
                yy1 = np.maximum(y1_[idx_self], y1_[idx_other])
                xx2 = np.minimum(x2_[idx_self], x2_[idx_other])
                yy2 = np.minimum(y2_[idx_self], y2_[idx_other])

                w_ = np.maximum(0.0, xx2 - xx1)
                h_ = np.maximum(0.0, yy2 - yy1)
                inter = w_ * h_

                if min_mode:
                    over = inter / np.minimum(areas[order[0]], areas[order[1:]])
                else:
                    over = inter / (areas[order[0]] + areas[order[1:]] - inter)

                inds = np.where(over <= nms_thresh)[0]
                order = order[inds + 1]

            return np.array(keep)

        # nc = prediction[0].shape[1] - 5  # number of classes
        xc = prediction[..., 4] > conf_thres  # candidates

        # Settings
        min_wh, max_wh = 2, 4096  # (pixels) minimum and maximum box width and height
        max_det = 300  # maximum number of detections per image
        time_limit = 10.0  # seconds to quit after

        t = time.time()
        output = [None] * prediction.shape[0]
        for xi, x in enumerate(prediction):  # image index, image inference
            # Apply constraints
            # x[((x[..., 2:4] < min_wh) | (x[..., 2:4] > max_wh)).any(1), 4] = 0  # width-height
            x = x[xc[xi]]  # confidence

            # If none remain process next image
            if not x.shape[0]:
                continue

            # Compute conf
            x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

            # Box (center x, center y, width, height) to (x1, y1, x2, y2)
            box_ = xywh2xyxy(x[:, :4])

            # i, j = (x[:, 5:] > conf_thres).nonzero(as_tuple=False).T
            i_, j = (x[:, 5:] > conf_thres).nonzero()
            # x = torch.cat((box[i], x[i, j + 5, None], j[:, None].float()), 1)
            x = np.array(np.concatenate((box_[i_], x[i_, j + 5, None], j[:, None]), 1)).astype(np.float32)

            # If none remain process next image
            n = x.shape[0]  # number of boxes
            if not n:
                continue

            # Sort by confidence
            # x = x[x[:, 4].argsort(descending=True)]

            # Batched NMS
            c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
            boxes, scores = x[:, :4] + c, x[:, 4]  # boxes (offset by class), scores
            i_ = nms_cpu(boxes, scores, iou_thres)
            if i_.shape[0] > max_det:  # limit detections
                i_ = i_[:max_det]

            output[xi] = x[i_]
            if (time.time() - t) > time_limit:
                break  # time limit exceeded

        return output

    def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None):

        def clip_coords(boxes, img_shape):
            boxes[:, 0].clip(0, img_shape[1])  # x1
            boxes[:, 1].clip(0, img_shape[0])  # y1
            boxes[:, 2].clip(0, img_shape[1])  # x2
            boxes[:, 3].clip(0, img_shape[0])  # y2

        # Rescale coords (xyxy) from img1_shape to img0_shape
        if ratio_pad is None:  # calculate from img0_shape
            gain = max(img1_shape) / max(img0_shape)  # gain  = old / new
            pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (
                    img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
        else:
            gain = ratio_pad[0][0]
            pad = ratio_pad[1]

        coords[:, [0, 2]] -= pad[0]  # x padding
        coords[:, [1, 3]] -= pad[1]  # y padding
        coords[:, :4] /= gain
        clip_coords(coords, img0_shape)
        return coords

    res = []
    image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    img_processed, w, h = image_preprocess(np.copy(image), [input_size, input_size])
    image_data = img_processed[np.newaxis, ...].astype(np.float32)
    image_data = np.transpose(image_data, [0, 3, 1, 2])

    inputs = [image_data]
    pred = onnx_helper.run(sess_tuple, inputs)[0]

    batch_detections = np.array(pred)
    batch_detections = non_max_suppression(batch_detections, conf_thres=0.4, iou_thres=0.5, agnostic=False)
    detections = batch_detections[0]
    if detections is not None:
        labels = detections[..., -1]
        boxs = detections[..., :4]
        confs = detections[..., 4]
        boxs[:, :] = scale_coords((input_size, input_size), boxs[:, :], (h, w)).round()
        for i, box in enumerate(boxs):
            x1, y1, x2, y2 = box
            class_name = class_names[int(labels[i])]
            score = confs[i]
            res.append({constants.RESULT_KEY_RECT: [y1, x1, y2, x2],
                        constants.RESULT_KEY_SCORE: score,
                        constants.RESULT_KEY_CLASS_NAME: class_name})

    # for i in range(len(out_boxes)):
    #    res.append({constants.RESULT_KEY_RECT: out_boxes[i], constants.RESULT_KEY_SCORE: out_scores[i], constants.RESULT_KEY_CLASS_NAME: out_classes[i]})
    return res


def predict_v4(sess_tuple, input_size, class_names, frame):
    h, w, _ = frame.shape

    def nms_cpu(boxes_, confs, nms_thresh=0.5, min_mode=False):
        x1 = boxes_[:, 0]
        y1 = boxes_[:, 1]
        x2 = boxes_[:, 2]
        y2 = boxes_[:, 3]

        areas = (x2 - x1) * (y2 - y1)
        order = confs.argsort()[::-1]

        keep = []
        while order.size > 0:
            idx_self = order[0]
            idx_other = order[1:]

            keep.append(idx_self)

            xx1 = np.maximum(x1[idx_self], x1[idx_other])
            yy1 = np.maximum(y1[idx_self], y1[idx_other])
            xx2 = np.minimum(x2[idx_self], x2[idx_other])
            yy2 = np.minimum(y2[idx_self], y2[idx_other])

            w_ = np.maximum(0.0, xx2 - xx1)
            h_ = np.maximum(0.0, yy2 - yy1)
            inter = w_ * h_

            if min_mode:
                over = inter / np.minimum(areas[order[0]], areas[order[1:]])
            else:
                over = inter / (areas[order[0]] + areas[order[1:]] - inter)

            inds = np.where(over <= nms_thresh)[0]
            order = order[inds + 1]

        return np.array(keep)

    def post_processing(conf_thresh, nms_thresh, output):

        # anchors = [12, 16, 19, 36, 40, 28, 36, 75, 76, 55, 72, 146, 142, 110, 192, 243, 459, 401]
        # num_anchors = 9
        # anchor_masks = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
        # strides = [8, 16, 32]
        # anchor_step = len(anchors) // num_anchors

        # [batch, num, 1, 4]
        box_array = output[0]
        # [batch, num, num_classes]
        confs = output[1]

        if type(box_array).__name__ != 'ndarray':
            box_array = box_array.cpu().detach().numpy()
            confs = confs.cpu().detach().numpy()

        num_classes = confs.shape[2]

        # [batch, num, 4]
        box_array = box_array[:, :, 0]

        # [batch, num, num_classes] --> [batch, num]
        max_conf = np.max(confs, axis=2)
        max_id = np.argmax(confs, axis=2)

        bboxes_batch = []
        for i_ in range(box_array.shape[0]):
            argwhere = max_conf[i_] > conf_thresh
            l_box_array = box_array[i_, argwhere, :]
            l_max_conf = max_conf[i_, argwhere]
            l_max_id = max_id[i_, argwhere]

            bboxes = []
            # nms for each class
            for j in range(num_classes):

                cls_argwhere = l_max_id == j
                ll_box_array = l_box_array[cls_argwhere, :]
                ll_max_conf = l_max_conf[cls_argwhere]
                ll_max_id = l_max_id[cls_argwhere]

                keep = nms_cpu(ll_box_array, ll_max_conf, nms_thresh)

                if keep.size > 0:
                    ll_box_array = ll_box_array[keep, :]
                    ll_max_conf = ll_max_conf[keep]
                    ll_max_id = ll_max_id[keep]

                    for k in range(ll_box_array.shape[0]):
                        bboxes.append(
                            [ll_box_array[k, 0], ll_box_array[k, 1], ll_box_array[k, 2], ll_box_array[k, 3],
                             ll_max_conf[k], ll_max_conf[k], ll_max_id[k]])

            bboxes_batch.append(bboxes)
        return bboxes_batch

    # IN_IMAGE_H = sess.get_inputs()[0].shape[2]
    # IN_IMAGE_W = sess.get_inputs()[0].shape[3]
    IN_IMAGE_H = IN_IMAGE_W = input_size

    # resized = cv2.resize(frame, (IN_IMAGE_W, IN_IMAGE_H), interpolation=cv2.INTER_LINEAR)
    resized = image_helper.resize_best_quality(frame, (IN_IMAGE_W, IN_IMAGE_H))
    img_in = cv2.cvtColor(resized, cv2.COLOR_BGR2RGB)
    img_in = np.transpose(img_in, (2, 0, 1)).astype(np.float32)
    img_in = np.expand_dims(img_in, axis=0)
    img_in /= 255.0

    # input_name = sess.get_inputs()[0].name
    # outputs = sess.run(None, {input_name: img_in})
    inputs = [img_in]
    outputs = onnx_helper.run(sess_tuple, inputs)

    boxes = post_processing(0.4, 0.6, outputs)

    def process_boxes(boxes_, width, height, class_names_):
        out_boxes1 = []
        out_scores1 = []
        out_classes1 = []
        for box in boxes_[0]:
            if len(box) >= 7:
                x1 = int(box[0] * width)
                y1 = int(box[1] * height)
                x2 = int(box[2] * width)
                y2 = int(box[3] * height)
                out_boxes1.append([y1, x1, y2, x2])
                out_scores1.append(box[5])
                out_classes1.append(class_names_[box[6]])
        return out_boxes1, out_scores1, out_classes1

    out_boxes, out_scores, out_classes = process_boxes(boxes, w, h, class_names)

    res = []
    for i in range(len(out_boxes)):
        res.append({constants.RESULT_KEY_RECT: out_boxes[i], constants.RESULT_KEY_SCORE: out_scores[i],
                    constants.RESULT_KEY_CLASS_NAME: out_classes[i]})
    return res


def yolo_label_maker(frame, rects, class_id, save_dir):
    lines = []
    for rect in rects:
        y1, x1, y2, x2 = rect
        h, w = image_helper.image_h_w(frame)
        cx = ((x1 + x2) * 0.5) / w
        cy = ((y1 + y2) * 0.5) / h
        w = (x2 - x1) / w
        h = (y2 - y1) / h
        line = "{} {} {} {} {}".format(class_id, cx, cy, w, h)
        lines.append(line)

    name = str(uuid.uuid4().hex)
    image_fn = file_helper.path_join(save_dir, name + ".jpg")
    label_fn = file_helper.path_join(save_dir, name + ".txt")
    cv2.imwrite(image_fn, frame)
    file_helper.write_lines(label_fn, lines)



from fnmatch import fnmatch


def wildcard(txt, pattern, case_insensitive=True):
    if txt == pattern:
        return True
    else:
        return fnmatch(txt.lower(), pattern.lower()) if case_insensitive else fnmatch(txt, pattern)


def wildcard_match_count(list_txt, pattern, case_insensitive=True):
    count = 0
    for txt in list_txt:
        if wildcard(txt, pattern, case_insensitive):
            count += 1
    return count


def wildcard_has_match(list_txt, pattern, case_insensitive=True):
    for txt in list_txt:
        if wildcard(txt, pattern, case_insensitive):
            return True
    return False



import os
import time

from ndu_gate_camera.utility import file_helper


def parse_class_names(classes_fn):
    return [line.rstrip('\n') for line in open(classes_fn, encoding='utf-8')]


_handler = None


def init(onnx_runner, log):
    global _handler
    if onnx_runner is None or onnx_runner == "onnxruntime":
        _handler = _OnnxruntimeHandler()
    elif onnx_runner == "openvino":
        _handler = _OpenvinoHandler()
    else:
        log.error("Bad onnx_runner: {}".format(onnx_runner))
        _handler = _OnnxruntimeHandler()


def get_sess_tuple(onnx_fn, output_names=None):
    return _handler.get_sess_tuple(onnx_fn, output_names)


def run(sess_tuple, inputs):
    return _handler.run(sess_tuple, inputs)


class _OnnxruntimeHandler:
    @staticmethod
    def get_sess_tuple(onnx_fn, output_names):
        import onnxruntime as rt
        start = time.time()
        sess = rt.InferenceSession(onnx_fn)
        elapsed = time.time() - start
        print("onnx model {} load time: {:.0f}sn".format(onnx_fn, elapsed))

        input_names = []
        for sess_input in sess.get_inputs():
            input_names.append(sess_input.name)

        if output_names is None:
            outputs = sess.get_outputs()
            output_names = []
            for output in outputs:
                output_names.append(output.name)

        return sess, input_names, output_names

    @staticmethod
    def run(sess_tuple, inputs):
        sess, input_names, output_names = sess_tuple

        if len(input_names) > 1:
            input_item = {}
            for i in range(len(inputs)):
                name = input_names[i]
                input_item[name] = inputs[i]
        else:
            input_item = {input_names[0]: inputs[0]}
        return sess.run(output_names, input_item)
        # res = sess.run(output_names, input_item)
        # return res


class _OpenvinoHandler:
    @staticmethod
    def get_sess_tuple(onnx_fn, output_names):
        from openvino.inference_engine import IECore, IENetwork

        ie = IECore()

        dir_name, name, extension = file_helper.get_file_name_extension(onnx_fn)
        xml_fn = file_helper.path_join(dir_name, name + ".xml")
        bin_fn = file_helper.path_join(dir_name, name + ".bin")
        if os.path.isfile(xml_fn) and os.path.isfile(bin_fn):
            # net = IENetwork(model=xml_fn, weights=bin_fn)
            net = IENetwork(xml_fn, bin_fn)
        else:
            net = ie.read_network(onnx_fn)

        # net = ie.read_network(onnx_fn)

        # sess = ie.load_network(network=net, device_name="CPU")
        sess = ie.load_network(net, "CPU")

        # exec_net = ie.load_network(network=net, device_name="CPU")
        # input_blob = next(iter(net.inputs))
        # outputs = net.outputs.keys()

        input_names = []
        for sess_input in sess.inputs:
            input_names.append(sess_input)

        if output_names is None:
            output_names = []
            for output in sess.outputs:
                output_names.append(output)

        # if not has_single_input:
        #     _input_item = {}
        #     for i in range(len(input_names)):
        #         name = input_names[i]
        #         _input_item[name] = None
        # else:
        #     _input_item = {input_names[0]: None}

        return sess, input_names, output_names

    @staticmethod
    def run(sess_tuple, inputs):
        sess, input_names, output_names = sess_tuple

        if len(input_names) > 1:
            input_item = {}
            for i in range(len(inputs)):
                name = input_names[i]
                input_item[name] = inputs[i]
        else:
            input_item = {input_names[0]: inputs[0]}

        pred = sess.infer(input_item)
        res = []
        for key, val in pred.items():
            if key in output_names:
                res.append(val)
        return res

        # pred = sess.infer(input_item)
        # return [pred["output"]]



DEFAULT_TB_GATEWAY_CONF = "/etc/thingsboard-gateway/config/tb_gateway.yaml"
DEFAULT_NDU_GATE_CONF = "/etc/ndu-gate/config/ndu_gate.yaml"
DEFAULT_NDU_GATE_CONF_WIN = "C:/ndu-gate/config/ndu_gate.yaml"

NDU_GATE_MODEL_FOLDER = "/etc/ndu-gate/data"

DEFAULT_HANDLER_SETTINGS = {
    "type": "SOCKET",
    "socket": {
        "port": 60060,
        "host": "127.0.0.1"
    }
}


RESULT_KEY_TRACK_ID = "track_id"
# [x1, y1, x2, y2]
RESULT_KEY_RECT = "rect"
RESULT_KEY_RECT_COLOR = "rect_color"
RESULT_KEY_RECT_DEBUG_TEXT = "rect_debug_text"
# string
RESULT_KEY_CLASS_NAME = "class_name"
# string - used for debug previewing
RESULT_KEY_PREVIEW_KEY = "preview_key"
# 0-1 arasında float olasılık değeri.
RESULT_KEY_SCORE = "score"
# Platform'a gönderilmesi istenen veri. İçeriği dictionary olmalıdır.
RESULT_KEY_DATA = "data"
# Preview'e büyük şekilde yazılacak veri.
RESULT_KEY_DEBUG = "debug"

EXTRA_DATA_KEY_RESULTS = "results"



from time import sleep
import cv2

from ndu_gate_camera.api.video_source import VideoSource, log


class CameraVideoSource(VideoSource):
    def __init__(self, source_config):
        super().__init__()
        self.__capture = self._find_vide_capture()
        self.__mirror = source_config.get("mirror", False)

    def get_frames(self):
        log.info("start camera streaming..")
        count = 0
        try:
            while self.__capture.isOpened():
                ok, frame = self.__capture.read()
                if not ok:
                    continue
                if self.__mirror:
                    frame = cv2.flip(frame, 1)
                count += 1
                yield count, frame
        except Exception as e:
            log.error(e)
        finally:
            log.info("camera stream finished..")
            self.__capture.release()
            cv2.destroyAllWindows()

    def reset(self):
        pass

    def stop(self):
        pass

    @staticmethod
    def _find_vide_capture():
        index = -1
        cap = cv2.VideoCapture(index)
        r, fr = cap.read()
        while fr is None:
            index += 1
            if index > 100:
                raise Exception('Cannot capture camera video')
            try:
                cap = cv2.VideoCapture(index)
                r, fr = cap.read()
            except:
                fr = None
        return cap



import time
from os import path
import cv2

from ndu_gate_camera.api.video_source import VideoSource, log
from ndu_gate_camera.utility import file_helper


class ImageVideoSource(VideoSource):
    def __init__(self, source_config):
        super().__init__()
        log.info("source_config %s", source_config)
        image_path = source_config.get("image_path", None)
        if image_path is None:
            raise ValueError("Image file path is empty")

        if path.isfile(image_path) is False:
            # TODO - get default installation path using OS
            image_path = file_helper.path_join(source_config.get("data_folder", "var/lib/ndu_gate_camera/data/".replace('/', path.sep)), image_path)
            if path.isfile(image_path) is False:
                dir_name = image_path[:image_path.rfind(path.sep)]
                file_name = image_path[image_path.rfind(path.sep) + len(path.sep):]
                self._images = []
                if path.isdir(dir_name):
                    for name in file_helper.enumerate_files(dir_name, False, file_name):
                        self._images.append(file_helper.path_join(dir_name, name))
            else:
                self._images = [image_path]
        else:
            self._images = [image_path]

        if len(self._images) == 0:
            raise ValueError("Image file is not exist : %s", image_path)

    def get_frames(self):
        log.debug("start image streaming..")
        count = 0
        i = 0
        start = time.time()
        while True:
            try:
                image_path = self._images[i]
                frame = cv2.imread(image_path)
                yield count, frame
                count += 1
            except:
                start = 0
            if time.time() - start > 5:
                start = time.time()
                i += 1
                if i == len(self._images):
                    i = 0

    def reset(self):
        pass

    def stop(self):
        # TODO
        pass



from os import path
import cv2

from ndu_gate_camera.api.video_source import VideoSource, log


class FileVideoSource(VideoSource):
    def __init__(self, source_config):
        """

        :param source_config:
        """
        super().__init__()
        log.info("source_config %s", source_config)
        self.__video_path = source_config.get("file_path", None)
        if self.__video_path is None:
            raise ValueError("Video file path is empty")

        if path.isfile(self.__video_path) is False:
            # TODO - get default installation path using OS
            self.__video_path = source_config.get("data_folder", "var/lib/ndu_gate_camera/data/".replace('/',
                                                                                                         path.sep)) + self.__video_path

        if path.isfile(self.__video_path) is False:
            raise ValueError("Video file is not exist : %s", self.__video_path)
        self._set_capture()

    def get_frames(self):
        log.debug("start video streaming..")
        count = 0
        # TODO - bitince başa sar?
        while self.__capture.isOpened():
            ret, frame = self.__capture.read()
            if ret is False:
                break

            yield count, frame
            count += 1

        log.debug("video finished..")
        self.__capture.release()
        cv2.destroyAllWindows()

        try:
            count = 0
            while self.__capture.isOpened():
                ok, frame = self.__capture.read()
                if not ok:
                    break
                yield count, frame
                count += 1
        except Exception as e:
            log.error(e)
        finally:
            log.debug("video finished..")
            self.__capture.release()
            cv2.destroyAllWindows()

    def reset(self):
        self._set_capture()

    def stop(self):
        # TODO
        pass

    def _set_capture(self):
        if path.isfile(self.__video_path) is False:
            raise ValueError("Video file is not exist")
        self.__capture = cv2.VideoCapture(self.__video_path)



from abc import ABC, abstractmethod

import logging

log = logging.getLogger("result")


class ResultHandler(ABC):

    @abstractmethod
    def save_result(self, result, device=None, runner_name=None, data_type='telemetry'):
        pass



from setuptools import setup

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

VERSION = "0.1.9.4"

setup(
    name='ndu_gate_camera',
    version=VERSION,
    license='Apache Software License (Apache Software License 2.0)',
    author='Netcad Innovation Labs @NETCAD',
    author_email='netcadinnovationlabs@gmail.com',
    description='NDU Gate Camera Service',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    python_requires=">=3.5",
    packages=['ndu_gate_camera', 'ndu_gate_camera.api', 'ndu_gate_camera.camera',
              'ndu_gate_camera.camera.video_sources', 'ndu_gate_camera.camera.result_handlers',
              'ndu_gate_camera.runners',
              'ndu_gate_camera.utility'],
    url='https://github.com/netcadlabs/ndu-gate',
    download_url='https://github.com/netcadlabs/ndu-gate/archive/%s.tar.gz' % VERSION,
    install_requires=[
        'pip',
        'gdown',
        'PyYAML',
        'simplejson',
        'vidgear',
        'pyzmq',
        'filterpy==1.4.5'
    ],
    package_data={
        "*": ["config/*"]
    },
    entry_points={
        'console_scripts': [
            'ndu-gate = ndu_gate_camera.ndu_camera:daemon'
        ]}
)



import os
import random
import time
from threading import Thread
from os import path
from logging import getLogger

from simplejson import load
import cv2
import numpy as np
from yaml import safe_load

from ndu_gate_camera.api.video_source import VideoSourceType
from ndu_gate_camera.camera.roi_manager import ROIManager
from ndu_gate_camera.camera.video_sources.camera_video_source import CameraVideoSource
from ndu_gate_camera.camera.video_sources.file_video_source import FileVideoSource
from ndu_gate_camera.camera.video_sources.ip_camera_video_source import IPCameraVideoSource
from ndu_gate_camera.camera.video_sources.pi_camera_video_source import PiCameraVideoSource
from ndu_gate_camera.camera.video_sources.youtube_video_source import YoutubeVideoSource
from ndu_gate_camera.camera.video_sources.image_video_source import ImageVideoSource
from ndu_gate_camera.utility import constants, image_helper, string_helper
from ndu_gate_camera.utility.ndu_utility import NDUUtility
from ndu_gate_camera.utility.download_zipped_helper import download_runner_data

DEFAULT_RUNNERS = {
    # "drivermonitor": "DriverMonitorRunner",
    # "socialdistance": "SocialDistanceRunner",
    # "emotionanalysis": "EmotionAnalysisRunner",
}

log = getLogger("service")


class NDUCameraService(Thread):
    def __init__(self, instance=None, config_dir="", handler=None, is_main_thread=True, extension_folder=None):
        self.__is_main_thread = is_main_thread
        if instance is None:
            instance = {}
        super().__init__()
        self.__result_handler = handler
        self._ndu_gate_config_dir = config_dir
        self.__extension_folder = extension_folder
        self.RUNNERS = instance.get("runners", [])
        self.SOURCE_TYPE = VideoSourceType.CAMERA
        self.SOURCE_CONFIG = None

        if instance.get("source"):
            self.SOURCE_CONFIG = instance.get("source")
            type_str = self.SOURCE_CONFIG.get("type", "PI_CAMERA")
            if VideoSourceType[type_str]:
                self.SOURCE_TYPE = VideoSourceType[type_str]
            else:
                self.SOURCE_TYPE = VideoSourceType.PI_CAMERA

        self.__frame_send_interval = self.SOURCE_CONFIG.get("frame_send_interval", 1000)
        self.__motion_kernel = self.SOURCE_CONFIG.get("motion_kernel", None)
        self.__camera_refresh_interval = self.SOURCE_CONFIG.get("camera_refresh_interval", 5000)
        self.__preview_show_motion_kernel = self.SOURCE_CONFIG.get("preview_show_motion_kernel", False)
        self.__preview_show = self.SOURCE_CONFIG.get("preview_show", False)
        if self.__preview_show:
            self.__preview_inited = False
            self.__last_preview_image = None
            self.__preview_show_debug_texts = self.SOURCE_CONFIG.get("preview_show_debug_texts", True)
            self.__preview_show_runner_info = self.SOURCE_CONFIG.get("preview_show_runner_info", True)
            self.__preview_show_score = self.SOURCE_CONFIG.get("preview_show_score", False)
            self.__preview_show_rect_name = self.SOURCE_CONFIG.get("preview_show_rect_name", True)
            self.__preview_show_track_id = self.SOURCE_CONFIG.get("preview_show_track_id", True)
            self.__preview_show_track_pnt = self.SOURCE_CONFIG.get("preview_show_track_pnt", False)
            if self.__preview_show_track_pnt:
                self._track_pnt_layer = None
                self._track_pnts = {}
            f = self.SOURCE_CONFIG.get("preview_show_rect_filter", None)
            if f is not None:
                if isinstance(f, str):
                    self.__preview_show_rect_filters = [f]
                elif isinstance(f, list):
                    self.__preview_show_rect_filters = f
                else:
                    raise Exception('Bad "preview_show_rect_filter" definition!')
            else:
                self.__preview_show_rect_filters = None

            self.__preview_write = self.SOURCE_CONFIG.get("preview_write", False)
            self.__preview_last_data = []
            self.__preview_last_data_show_times = []
            self.__preview_write_file_name = self.SOURCE_CONFIG.get("preview_write_file_name", "")
            self.__total_elapsed_times = []

        self._track_history = {}
        self._exit_requested = False
        self.__max_frame_dim = self.SOURCE_CONFIG.get("max_frame_dim", None)
        self.__min_frame_dim = self.SOURCE_CONFIG.get("min_frame_dim", None)
        self.__sleep = self.SOURCE_CONFIG.get("sleep", 0)
        self.__skip_frame = self.SOURCE_CONFIG.get("skip_frame", 0)
        self._skip = 0
        self.__color_toggle = None

        self.frame_sent = self.SOURCE_CONFIG.get("frame_sent", False)

        self.default_runners = DEFAULT_RUNNERS
        self.runners_configs = []
        self.runners_configs_by_key = {}
        self.implemented_runners = {}
        self.available_runners = {}
        self.download_urls = {}

        self._load_download_urls()
        self._load_runners()
        self._connect_with_runners()

        self.video_source = None
        self._steps = {}
        # self._start()

    # def run(self):
    #     self.start()

    def _load_download_urls(self, ):
        load_download_urls = path.join(self._ndu_gate_config_dir, 'data_download_urls.yaml')
        print("_load_download_urls config file : {}".format(load_download_urls))
        try:
            with open(load_download_urls, encoding="utf-8") as general_config:
                self.download_urls = safe_load(general_config)

        except:
            print('TODO - print exception details')

    def _load_runners(self):
        """
        config dosyasında belirtilen NDUCameraRunner imaplementasyonlarını bulur ve _implemented_runners içerisine ekler
        Aynı şekilde herbir runner için config dosyalarını bulur ve connectors_configs içerisine ekler
        """
        runners_configs_temp = {}
        last_priority = 1000000

        if self.RUNNERS:
            for runner in self.RUNNERS:
                log.debug("runner config : %s", runner)
                try:
                    if runner.get("status", 1) == 0:  # runner status default value is 1>>>
                        log.debug("runner is not active %s", runner)
                        continue

                    runner_type = runner.get("type", None)
                    if runner_type is None:
                        log.warning("type not found for %s", runner)
                        continue

                    class_name = self.default_runners.get(runner_type, runner.get("class", None))
                    if class_name is None:
                        log.warning("class name not found for %s", runner)
                        continue

                    runner_class_info = NDUUtility.check_and_import(runner_type, class_name,
                                                                    package_uuids=runner.get("uuids", None),
                                                                    extension_folder=self.__extension_folder)

                    if runner_class_info is None or runner_class_info[0] is None:
                        log.warning("class name implementation not found for %s - %s", runner_type, class_name)
                        continue

                    runner_class = runner_class_info[0]

                    runner_key = self.__get_runner_key(runner_type, class_name)
                    self.implemented_runners[runner_key] = runner_class

                    configuration_name = runner['configuration']
                    config_file = self._ndu_gate_config_dir + configuration_name

                    if path.isfile(config_file):
                        with open(config_file, 'r', encoding="UTF-8") as conf_file:
                            runner_conf = load(conf_file)
                            runner_conf["name"] = runner["name"]
                    else:
                        log.error("config file is not found %s", config_file)
                        runner_conf = {"name": runner["name"]}

                    roi_manager = None
                    roi_configuration_name = runner.get('roi_configuration', None)
                    if roi_configuration_name is not None:
                        roi_config_file = self._ndu_gate_config_dir + roi_configuration_name
                        try:
                            if path.isfile(roi_config_file):
                                with open(roi_config_file, 'r', encoding="UTF-8") as roi_conf_file:
                                    runner_roi_conf = load(roi_conf_file)
                                    roi_manager = ROIManager(runner_roi_conf, runner["name"])
                                    runner_conf["roi_manager"] = roi_manager
                            else:
                                log.error("roi config file is not found %s", roi_config_file)
                        except Exception as e1:
                            log.error("Error on loading runner roi config %s", roi_config_file)
                            log.exception(e1)

                    runner_custom_conf = {}
                    custom_config_file = self._ndu_gate_config_dir + runner_type + "_custom.json"
                    if path.isfile(custom_config_file):
                        with open(custom_config_file, 'r', encoding="UTF-8") as conf_file:
                            runner_custom_conf = load(conf_file)
                    else:
                        log.error("custom config file is not found %s", custom_config_file)

                    runner_conf["custom_config"] = runner_custom_conf

                    runner_unique_key = self.__get_runner_configuration_key(runner_type, class_name, configuration_name)

                    runner_priority = runner.get("priority", None)
                    if runner_priority is None:
                        runner_priority = last_priority
                        last_priority = last_priority + 100

                    rt = runners_configs_temp[runner_unique_key] = {
                        "name": runner["name"],
                        "type": runner_type,
                        "class": runner_class,
                        "configuration_name": configuration_name,
                        "config": runner_conf,
                        "priority": runner_priority,
                        "runner_key": runner_key,
                        "runner_unique_key": runner_unique_key,
                        "step_frame": runner.get("step_frame", None),
                        "step_sec": runner.get("step_sec", None),
                        "runner_load_path": runner_class_info[1]
                    }
                    if roi_manager is not None:
                        rt["roi_manager"] = roi_manager

                except Exception as e:
                    log.error("Error on loading runner config")
                    log.exception(e)

            runner_arr = []
            # add all configs to array
            for key in runners_configs_temp:
                runner_arr.append(runners_configs_temp[key])

            self.runners_configs_by_key = runners_configs_temp
            self.runners_configs = sorted(runner_arr, key=lambda x: x["priority"], reverse=False)
        else:
            log.error("Runners - not found! Check your configuration!")

    def _connect_with_runners(self):
        """
        runners_configs içindeki configleri kullanarak sırayla yüklenen runner sınıflarının instance'larını oluşturur
        oluşturulan bu nesneleri available_runners içerisine ekler.
        """
        for runner_config in self.runners_configs:
            runner = None
            try:
                runner_key = runner_config["runner_key"]
                runner_type = runner_config["type"]
                runner_unique_key = runner_config["runner_unique_key"]

                if runner_config["config"] is None:
                    log.warning("Config not found for %s", runner_key)
                    continue

                # data klasörünün hangi runner klasörüne indirileceği problemi var, şimdilik projedeki runners klasörüne indirsin
                download_runner_data(runner_type, runner_config["runner_load_path"], self.download_urls)

                if self.implemented_runners[runner_key] is None:
                    log.error("Implemented runner not found for %s", runner_key)
                else:
                    runner = self.implemented_runners[runner_key](runner_config["config"], runner_type)
                    # runner.setName(runner_config["name"])
                    self.available_runners[runner_unique_key] = runner

            except Exception as e:
                log.exception(e)
                if runner is not None and NDUUtility.has_method(runner, 'close'):
                    runner.close()

    def _set_video_source(self):
        """
        SOURCE_TYPE değerine göre video_source değişkenini oluşturur.
        """
        try:
            if self.SOURCE_TYPE is VideoSourceType.VIDEO_FILE:
                self.SOURCE_CONFIG["test_data_path"] = path.dirname(
                    path.dirname(path.abspath(__file__))) + '/data/'.replace('/', path.sep)
                self.video_source = FileVideoSource(self.SOURCE_CONFIG)
                pass
            elif self.SOURCE_TYPE is VideoSourceType.PI_CAMERA:
                self.video_source = PiCameraVideoSource(show_preview=True)
            elif self.SOURCE_TYPE is VideoSourceType.VIDEO_URL:
                # TODO
                pass
            elif self.SOURCE_TYPE is VideoSourceType.IP_CAMERA:
                self.video_source = IPCameraVideoSource(self.SOURCE_CONFIG)
                pass
            elif self.SOURCE_TYPE is VideoSourceType.CAMERA:
                self.video_source = CameraVideoSource(self.SOURCE_CONFIG)
            elif self.SOURCE_TYPE is VideoSourceType.YOUTUBE:
                self.video_source = YoutubeVideoSource(self.SOURCE_CONFIG)
            elif self.SOURCE_TYPE is VideoSourceType.IMAGE_FILE:
                self.video_source = ImageVideoSource(self.SOURCE_CONFIG)
            else:
                log.error("Video source type is not supported : %s ", self.SOURCE_TYPE.value)
                # exit(101)
        except Exception as e:
            log.error("Error during setting up video source")
            log.error(e)

    def exit_signal(self):
        self._exit_requested = True

    # main thread tarafından çağırılabilir
    def check_for_preview(self):
        if self.__is_main_thread:
            raise Exception('Main thread does not need to call check_for_preview!')
        if self.__preview_show:
            return self._show_preview()

    def _show_preview(self):
        if not self.__preview_inited:
            self._winname = self.SOURCE_CONFIG.get("device", "ndu_gate_camera preview")
            pos = self.SOURCE_CONFIG.get("preview_show_win_pos", [40, 30])
            self._pause = False
            cv2.namedWindow(self._winname)  # Create a named window
            cv2.moveWindow(self._winname, pos[0], pos[1])
            self.__preview_inited = True

        preview = self.__last_preview_image
        if preview is not None:
            self.__last_preview_image = None
            cv2.imshow(self._winname, preview)
            # while True:
            #     k = cv2.waitKey(100) & 0xFF
            #     print(k)
            while True:
                k = cv2.waitKey(1) & 0xFF
                if k == ord("q"):
                    self._exit_requested = True
                    break
                elif k == ord("s"):
                    self._skip = 30
                elif k == 32:  # space key
                    self._pause = not self._pause
                if not self._pause:
                    break
            if self.__preview_write:
                self._write_frame(preview)

    def finish_preview(self):
        if self.__preview_show and self.__preview_write and self.__out is not None:
            self.__out.release()
            import ffmpeg  # pip3 install ffmpeg-python & brew install ffmpeg
            try:
                # daha az sıkışmış, quicktime çalabiliyor
                fn = self.__preview_write_file_name
                ffmpeg.input(fn).output(fn + '_ffmpeg.mp4').run(capture_stdout=True, capture_stderr=True)

                # # # süper sıkışmış ama quicktime çalamıyor. Benim denemelerimde yarım yamalak kaydedebildi!
                # ffmpeg.input(fn) \
                #     .output(fn + '2.mp4', vcodec='libx265', crf=24, t=5) \
                #     .run(capture_stdout=True, capture_stderr=True)
                # # os.remove(fn)
            except ffmpeg.Error as e:
                print('stdout:', e.stdout.decode('utf8'))
                print('stderr:', e.stderr.decode('utf8'))
                raise e
            finally:
                self.__out = None

    def run(self):
        while not self._exit_requested:
            print("camera refreshing...")
            log.debug("camera refreshing...")
            self._run()

    def _run(self):
        self._set_video_source()
        if self.video_source is None:
            log.error("video source is not set!")
            return
            # exit(102)
        start_total = None
        # TODO - çalıştırma sırasına göre sonuçlar bir sonraki runnera aktarılabilir
        # TODO - runner dependency ile kimin çıktısı kimn giridisi olacak şeklinde de olabilir

        try:
            for runner_unique_key in self.available_runners:
                runner_conf = self.runners_configs_by_key[runner_unique_key]
                step_sec = runner_conf.get("step_sec", None)
                step_frame = runner_conf.get("step_frame", None)
                if step_sec is not None or step_frame is not None:
                    self._steps[runner_unique_key] = [step_sec, 0, step_frame]

            device = self.SOURCE_CONFIG.get("device", None)
            i = -1
            last_thresh = last_gray = None
            has_motion_kernel = self.__motion_kernel is not None
            results_for_preview = None
            for _frame_index, frame in self.video_source.get_frames():
                if self._exit_requested:
                    break
                i += 1
                if i > self.__camera_refresh_interval:
                    break
                if self.__skip_frame > 1 and i % self.__skip_frame != 0:
                    continue
                if has_motion_kernel:
                    gray = image_helper.resize(frame, width=500, interpolation=cv2.INTER_NEAREST)
                    gray = cv2.cvtColor(gray, cv2.COLOR_BGR2GRAY)
                    gray = cv2.GaussianBlur(gray, (self.__motion_kernel, self.__motion_kernel), 0)
                    if last_gray is not None:
                        frame_delta = cv2.absdiff(last_gray, gray)
                        last_thresh = cv2.threshold(frame_delta, 25, 255, cv2.THRESH_BINARY)[1]
                        no_zero = cv2.countNonZero(last_thresh)
                        if self.__preview_show_motion_kernel:
                            image_helper.put_text(last_thresh, str(no_zero), [50, 50])
                        # cv2.imshow("last_thresh", last_thresh)
                        # cv2.waitKey(1)
                        if no_zero <= 1:
                            continue
                    last_gray = gray

                if i % 500 == 0:
                    log.debug("frame count %s ", i)
                    print("Source Device : {} - frame {}".format(device, i))

                if i % self.__frame_send_interval == 0:
                    try:
                        camera_capture_base64 = image_helper.frame2base64(frame)
                        log.info("CAMERA_CAPTURE size : %s", len(camera_capture_base64))
                        print("CAMERA_CAPTURE size : {}".format(len(camera_capture_base64)))
                        if self.frame_sent:
                            if len(camera_capture_base64) < 16000:
                                self.__result_handler.save_result([{"data": {"CAMERA_CAPTURE": camera_capture_base64}}],
                                                                  device=device, data_type='attribute')
                        self.__result_handler.save_result([{"data": {"FRAME_COUNT": i}}], device=device,
                                                          data_type='attribute')
                        self.frame_sent = True
                    except Exception as e:
                        log.exception(e)
                        log.error("can not create CAMERA_CAPTURE")

                if self.__max_frame_dim is not None:
                    frame = image_helper.resize_if_larger(frame, self.__max_frame_dim)
                if self.__min_frame_dim is not None:
                    frame = image_helper.resize_if_smaller(frame, self.__min_frame_dim)

                if self.__preview_show:
                    results_for_preview = []
                if self._skip <= 0:
                    if self.__preview_show:
                        start_total = time.time()
                    extra_data = {
                        constants.EXTRA_DATA_KEY_RESULTS: {}
                    }
                    extra_data_results = extra_data[constants.EXTRA_DATA_KEY_RESULTS]
                    empty_resulted_runners = []
                    # TODO - check runner settings before send the frame to runner
                    for runner_unique_key in self.available_runners:
                        try:
                            start_time = time.time()
                            step = self._steps.get(runner_unique_key, None)
                            if step is not None:
                                step_sec = step[0]
                                last_time = step[1]
                                step_frame = step[2]
                                if step_frame is not None and i % step_frame != 0:
                                    continue
                                if step_sec is not None:
                                    diff = (start_time - last_time)
                                    if diff <= step_sec:
                                        continue
                                    else:
                                        step[1] = start_time

                            runner_conf = self.runners_configs_by_key[runner_unique_key]
                            if "roi_manager" in runner_conf:
                                roi_manager = runner_conf.get("roi_manager", None)
                                frame = roi_manager.forward(frame)
                                results = self.available_runners[runner_unique_key].process_frame(frame,
                                                                                                  extra_data=extra_data)
                                frame, results = roi_manager.reverse(frame, results)
                            else:
                                results = self.available_runners[runner_unique_key].process_frame(frame,
                                                                                                  extra_data=extra_data)

                            if results is not None and len(results) > 0:
                                extra_data_results[runner_unique_key] = results
                                log.debug("result : %s", results)
                            else:
                                empty_resulted_runners.append(runner_unique_key)
                            if self.__preview_show:
                                elapsed = time.time() - start_time
                                results_for_preview.append({"elapsed_time": '{}: {:.4f}sn fps:{:.0f}'.format(
                                    runner_conf["type"], elapsed, 1.0 / max(elapsed, 0.001))})

                        except Exception as e:
                            log.exception(e)

                    def check_results(runner_unique_key_, results_, device_, results_for_preview_):
                        self.__result_handler.save_result(results_, device=device_,
                                                          runner_name=self.runners_configs_by_key[runner_unique_key_][
                                                              "name"])
                        if self.__preview_show:
                            results_for_preview_.extend(results_)

                    delete_required_keys = []
                    for runner_unique_key, results in extra_data_results.items():
                        results = self._check_track(runner_unique_key, results)
                        if results is not None and len(results) > 0:
                            check_results(runner_unique_key, results, device, results_for_preview)
                        else:
                            delete_required_keys.append(runner_unique_key)
                    for runner_unique_key in delete_required_keys:
                        del extra_data_results[runner_unique_key]
                    for runner_unique_key in empty_resulted_runners:
                        results = self._check_track(runner_unique_key, None)
                        if results is not None and len(results) > 0:
                            extra_data_results[runner_unique_key] = results
                            check_results(runner_unique_key, results, device, results_for_preview)

                self._age_track_history()
                if self.__preview_show:
                    if self._skip > 0:
                        self._skip -= 1
                        preview = frame
                    else:
                        total_elapsed_time = time.time() - start_total
                        self.__total_elapsed_times.append(total_elapsed_time)
                        len_times = len(self.__total_elapsed_times)
                        total_elapsed_time = sum(self.__total_elapsed_times) / len_times
                        if len_times > 10000:
                            self.__total_elapsed_times = []
                        results_for_preview.append({"total_elapsed_time": '{:.0f}msec fps:{:.0f}'.format(
                            total_elapsed_time * 1000, (1.0 / max(total_elapsed_time, 0.001)))})
                        preview = self._get_preview(frame, results_for_preview)
                        if self.__preview_show_motion_kernel and last_thresh is not None:
                            th = image_helper.resize(last_thresh, preview.shape[1], preview.shape[0],
                                                     interpolation=cv2.INTER_NEAREST)
                            th = cv2.cvtColor(th, cv2.COLOR_GRAY2BGR)
                            preview = cv2.addWeighted(preview, 1, th, 1, 0)

                    self.__last_preview_image = preview
                    if self.__is_main_thread:
                        self._show_preview()

                    # # write frames
                    # fn = "C:/_koray/temp/sc19_{}.jpg".format(str(i))
                    # cv2.imwrite(fn, frame)

                if self.__sleep > 0 and self._skip <= 0:
                    time.sleep(self.__sleep)

            if self.__preview_show and self.__is_main_thread:
                self.finish_preview()

            # TODO - set camera_perspective
        except Exception as e:
            log.exception(e)

        log.info("Video source is finished")

    def _check_track(self, runner_unique_key, results):
        res = []
        if results is not None:
            for result in results:
                if constants.RESULT_KEY_TRACK_ID in result:
                    if runner_unique_key not in self._track_history:
                        self._track_history[runner_unique_key] = {}
                    history = self._track_history[runner_unique_key]
                    track_id = result[constants.RESULT_KEY_TRACK_ID]
                    if track_id != -1:
                        if track_id not in history:
                            history[track_id] = {"result": result, "age": 0}
                        else:
                            result0 = history[track_id]["result"]
                            history[track_id]["age"] = 0
                            score = result.get(constants.RESULT_KEY_SCORE, 0.9)
                            score0 = result0.get(constants.RESULT_KEY_SCORE, 0.9)
                            if score < score0:
                                for key in [constants.RESULT_KEY_DATA, constants.RESULT_KEY_CLASS_NAME,
                                            constants.RESULT_KEY_DEBUG]:
                                    if key in result0:
                                        result[key] = result0[key]
                                result[constants.RESULT_KEY_SCORE] = score0
                            else:
                                history[track_id]["result"] = result
                        if constants.RESULT_KEY_DATA in result:
                            result = result.copy()
                            del result[constants.RESULT_KEY_DATA]
                    elif constants.RESULT_KEY_DATA in result:
                        result = result.copy()
                        del result[constants.RESULT_KEY_DATA]
                    res.append(result)
                else:
                    res.append(result)

        for i, (runner_unique_key, history) in enumerate(self._track_history.items()):
            del_lst = []
            for track_id, h in history.items():
                if h["age"] > 30:
                    result = history[track_id]["result"]
                    res.append(result)
                    del_lst.append(track_id)
            for track_id in del_lst:
                del history[track_id]
        return res

    def _age_track_history(self):
        for i, (runner_unique_key, history) in enumerate(self._track_history.items()):
            for track_id, h in history.items():
                h["age"] += 1

    def _write_frame(self, frame):
        def get_free_file_name(fn):
            if os.path.exists(fn):
                filename, file_extension = os.path.splitext(fn)
                i = 1
                while os.path.exists(fn):
                    # sil fn = f"{filename}{i}{file_extension}"
                    fn = "{}{}{}".format(filename, i, file_extension)
                    i += 1
            return fn

        try:
            self.__out.write(frame)
        except:
            shape = frame.shape[1], frame.shape[0]
            fourcc = cv2.VideoWriter_fourcc(*'MP4V')
            # fourcc = cv2.VideoWriter_fourcc(*'AVC1')
            # fourcc = cv2.VideoWriter_fourcc(*'XVID')
            self.__preview_write_file_name = get_free_file_name(self.__preview_write_file_name)
            self.__out = cv2.VideoWriter(self.__preview_write_file_name, fourcc, 24.0, shape)
            self.__out.write(frame)

    def _new_color(self):
        if self.__color_toggle is None:
            self.__color_toggle = [random.randint(50, 255), random.randint(50, 255), random.randint(50, 255)]
            return self.__color_toggle
        else:
            c = self.__color_toggle
            self.__color_toggle = None
            return [max(50, 255 - c[0]), max(50, 255 - c[1]), max(50, 255 - c[2])]

    def _get_preview(self, image, results):
        def get_color(obj, class_preview_key_, color=None):
            if color is None:
                color = [255, 255, 255]
                if class_preview_key_ is not None:
                    if not hasattr(obj, "__colors"):
                        setattr(obj, "__colors", {})
                    dic = getattr(obj, "__colors")
                    if class_preview_key_ in dic:
                        color = dic[class_preview_key_]
                    else:
                        color = dic[class_preview_key_] = self._new_color()
            return color

        def draw_rect(obj, img, c1_, c2_, class_preview_key_, color=None):
            if color is None:
                color = get_color(obj, class_preview_key_, color)
            cv2.rectangle(img, (c1_[0], c1_[1]), (c2_[0], c2_[1]), color=[1, 1, 1], thickness=3)
            cv2.rectangle(img, (c1_[0], c1_[1]), (c2_[0], c2_[1]), color=color, thickness=2)

        show_debug_texts = self.__preview_show_debug_texts
        show_runner_info = self.__preview_show_runner_info
        show_score = self.__preview_show_score
        rect_filters = self.__preview_show_rect_filters
        show_rect_name = self.__preview_show_rect_name
        show_track_id = self.__preview_show_track_id
        show_track_pnt = self.__preview_show_track_pnt

        h, w, *_ = image.shape
        line_height = 20
        font_scale = 0.4
        current_line = [20, 0]
        current_line_bottom = [20, h]
        data_added = []
        if results is not None:
            for result in results:
                debug_texts = []
                text_type = ""
                has_data = False

                if show_runner_info:
                    total_elapsed_time = result.get("total_elapsed_time", None)
                    if total_elapsed_time is not None:
                        image_helper.put_text(image, total_elapsed_time, [w - 150, h - 30], color=[200, 200, 128],
                                              font_scale=0.4)
                        continue

                values = {}
                elapsed_time = values["elapsed_time"] = result.get("elapsed_time", None)
                class_name = values[constants.RESULT_KEY_CLASS_NAME] = result.get(constants.RESULT_KEY_CLASS_NAME, None)
                class_preview_key = values[constants.RESULT_KEY_PREVIEW_KEY] = result.get(
                    constants.RESULT_KEY_PREVIEW_KEY, class_name)
                score = values[constants.RESULT_KEY_SCORE] = result.get(constants.RESULT_KEY_SCORE, None)
                rect = values[constants.RESULT_KEY_RECT] = result.get(constants.RESULT_KEY_RECT, None)
                rect_debug_text = values[constants.RESULT_KEY_RECT_DEBUG_TEXT] = result.get(
                    constants.RESULT_KEY_RECT_DEBUG_TEXT, None)
                data = values[constants.RESULT_KEY_DATA] = result.get(constants.RESULT_KEY_DATA, None)
                debug_text = values[constants.RESULT_KEY_DEBUG] = result.get(constants.RESULT_KEY_DEBUG, None)

                if debug_text is not None:
                    debug_texts.append(debug_text)

                def _get_score_txt(text_, score_):
                    res = "%{:.2f} ".format(score_ * 100) if isinstance(score_, float) else score_
                    if text_ is not None and len(text_) > 0:
                        return "{}- {} ".format(text_, res)
                    else:
                        return res

                text = ""
                if class_name is not None:
                    text = NDUUtility.debug_conv_turkish(class_name) + " "
                    if show_score and score is not None:
                        text = _get_score_txt(text, score)
                elif show_score and score is not None:
                    text = _get_score_txt(None, score)

                if data is not None:
                    add_txt = " data: " + str(data)
                    data_added.append(add_txt)
                    text = text + add_txt
                    has_data = True
                if rect is not None:
                    c = np.array(rect[:4], dtype=np.int32)
                    c1, c2 = [c[1], c[0]], (c[3], c[2])
                    color_rect = get_color(self, class_preview_key, result.get(constants.RESULT_KEY_RECT_COLOR, None))
                    show_rect = True
                    if rect_filters is not None:
                        show_rect = False
                        for rect_filter in rect_filters:
                            if string_helper.wildcard(class_name, rect_filter):
                                show_rect = True
                                break
                    if show_rect:
                        draw_rect(self, image, c1, c2, class_preview_key, color_rect)
                    else:
                        text = ""
                    if show_rect and show_rect_name and len(text) > 0:
                        # c1[1] = c1[1] + line_height
                        c1[1] -= 10
                        if rect_debug_text is not None:
                            text += " - " + rect_debug_text
                        image_helper.put_text(image, text, c1)
                        text = ""
                    if show_rect and show_track_id:
                        track_id = result.get(constants.RESULT_KEY_TRACK_ID, None)
                        if track_id is not None:
                            # c1[1] -= line_height * 1.5
                            c1[1] -= 15
                            image_helper.put_text(image, "{}".format(track_id), c1)
                    if show_track_pnt:
                        track_id = result.get(constants.RESULT_KEY_TRACK_ID, None)
                        if track_id is not None:
                            if self._track_pnt_layer is None:
                                h, w = image_helper.image_h_w(image)
                                self._track_pnt_layer = np.zeros((h, w, 3), np.uint8)
                            self._track_pnt_layer = image_helper.change_brightness(self._track_pnt_layer, -1)
                            y1, x1, y2, x2 = tuple(rect)
                            pnt = (int(x1 + (x2 - x1) * 0.5), int(y2))
                            pnts_key = str(track_id)
                            pnts = self._track_pnts.get(pnts_key, [])
                            pnts.append(pnt)
                            self._track_pnts[pnts_key] = pnts
                            if len(pnts) > 1:
                                pts = np.array(pnts, np.int32)
                                # cv2.polylines(self._track_pnt_layer, [pts], False, color, 5)
                                # cv2.polylines(image, [pts], False, color_rect, 5)
                                thickness = 10
                                for i in range(len(pnts) - 1, 0, -1):
                                    p0 = pnts[i]
                                    p1 = pnts[i - 1]
                                    cv2.line(image, p0, p1, color_rect, thickness)
                                    if i % 10 == 0:
                                        thickness -= 1
                                        if thickness < 1:
                                            # pnts = pnts[i: len(pnts)]
                                            break

                            # color = [255,255,255]
                            # cv2.circle(self._track_pnt_layer, pnt, 1, color, 8)
                            # image = cv2.addWeighted(image, 1, self._track_pnt_layer, 1,)
                            # ret, mask = cv2.threshold(self._track_pnt_layer, 1, 255, cv2.THRESH_BINARY)
                            # mask = cv2.bitwise_not(mask)
                            # mask = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
                            # image = cv2.bitwise_and(image, image, mask=mask)
                            # image = cv2.addWeighted(image, 1, self._track_pnt_layer, 0.1, 1)
                            # image += self._track_pnt_layer
                            # image = self._track_pnt_layer.copy()

                if elapsed_time is not None:
                    text_type = elapsed_time + " " + text_type
                if len(text) > 0:
                    text_type = text_type + text + " "
                if show_runner_info and len(text_type) > 0:
                    current_line[1] += line_height
                    if show_runner_info:
                        if not has_data:
                            image_helper.put_text(image, text_type, current_line, font_scale=font_scale)
                        else:
                            image_helper.put_text(image, text_type, current_line, color=[0, 0, 255],
                                                  font_scale=font_scale)
                if show_debug_texts and len(debug_texts) > 0:
                    for debug_text in debug_texts:
                        current_line_bottom[1] -= line_height * 2
                        image_helper.put_text(image, debug_text, current_line_bottom, color=[255, 250, 99],
                                              font_scale=font_scale * 2.75, thickness=2, back_color=[1, 1, 1])

        if show_runner_info:
            if len(data_added) > 0:
                show_last_data_frame_time = time.time()
                for data in data_added:
                    if data not in self.__preview_last_data:
                        self.__preview_last_data.append(data)
                        self.__preview_last_data_show_times.append(show_last_data_frame_time)
                    else:
                        self.__preview_last_data_show_times[
                            self.__preview_last_data.index(data)] = show_last_data_frame_time
            for i in reversed(range(len(self.__preview_last_data))):
                elapsed = time.time() - self.__preview_last_data_show_times[i]
                if elapsed > 5:
                    del self.__preview_last_data_show_times[i]
                    del self.__preview_last_data[i]
            for last_data in self.__preview_last_data:
                current_line[1] += line_height
                image_helper.put_text(image, last_data, current_line, color=[0, 255, 255])

        return image

    @staticmethod
    def __get_runner_configuration_key(runner_type, class_name, configuration):
        if configuration is None:
            configuration = runner_type + ".json"
        return runner_type + "_" + class_name + "_" + configuration

    @staticmethod
    def __get_runner_key(type, class_name):
        return type + "_" + class_name


if __name__ == '__main__':
    NDUCameraService(constants.DEFAULT_NDU_GATE_CONF.replace('/', path.sep))



import codecs
import io
import os
import pathlib
from typing import AnyStr, Tuple
from shutil import copyfile

from ndu_gate_camera.utility import string_helper
from ndu_gate_camera.utility.ndu_utility import NDUUtility


def read_lines(filename, encoding="utf-8"):
    with io.open(filename, mode="r", encoding=encoding) as f:
        content = f.readlines()
        content = [x.strip() for x in content]
        return content


def write_all(filename, text, encoding="utf-8"):
    write_lines(filename, [text], encoding)


def write_lines(filename, lines, encoding="utf-8"):
    with codecs.open(filename, "w", encoding) as f:
        for item in lines:
            f.write("%s\n" % item)


def enumerate_files(dir_path, recursive=False, wildcard_pattern=None, case_insensitive=True):
    # if not recursive:
    #     from os import listdir
    #     from os.path import isfile, join
    #     if filter is None:
    #         for file_name_with_extension in [f for f in listdir(dir_path) if isfile(join(dir_path, f))]:
    #             yield os.path.join(dir_path, file_name_with_extension), file_name_with_extension
    #     else:
    #         for file_name_with_extension in [f for f in listdir(dir_path) if isfile(join(dir_path, f))]:
    #             if string_helper.wildcard(file_name_with_extension, filter, case_insensitive):
    #                 yield os.path.join(dir_path, file_name_with_extension), file_name_with_extension
    if wildcard_pattern is None:
        for root, sub_dirs, files in os.walk(dir_path):
            for name in files:
                yield path_join(root, name)
            if not recursive:
                break
    else:
        for root, sub_dirs, files in os.walk(dir_path):
            for name in files:
                # name = os.path.basename(fn)
                if string_helper.wildcard(name, wildcard_pattern, case_insensitive=case_insensitive):
                    yield path_join(root, name)
            if not recursive:
                break


def append_line(filename, line, encoding="utf-8"):
    with codecs.open(filename, "a", encoding) as f:
        f.write("%s\n" % line)


def get_parent_dir_path(file_name):
    return str(pathlib.Path(file_name).parent.absolute())


def get_parent_dir_name(file_name):
    return str(pathlib.Path(file_name).parent.name)


def path_join(a: AnyStr, *paths: AnyStr) -> AnyStr:
    return os.path.join(a, *paths).replace("/", os.path.sep)


def create_dir(dir_name, parents=True, exist_ok=True):
    from pathlib import Path
    Path(dir_name).mkdir(parents=parents, exist_ok=exist_ok)


def copy_file(source, target):
    copyfile(source, target)


def get_file_name_extension(file_full_name) -> Tuple[AnyStr, AnyStr, AnyStr]:
    dir_name, file_name = os.path.split(file_full_name)
    name, extension = os.path.splitext(file_name)
    return dir_name, name, extension


def delete_file(file_name):
    os.remove(file_name)